__all__ = [
    "FloquetBasis",
    "floquet_tensor",
    "fsesolve",
    "fmmesolve",
    "FMESolver",
]

import numpy as np
from qutip.core import data as _data
from qutip.core.data import Data
from qutip import Qobj, QobjEvo
from .propagator import Propagator
from .mesolve import MESolver
from .solver_base import Solver
from .integrator import Integrator
from .result import Result
from time import time
from ..ui.progressbar import progress_bars


class FloquetBasis:
    """
    Utility to compute floquet modes and states.

    Attributes
    ----------
    U : :class:`.Propagator`
        The propagator of the Hamiltonian over one period.

    evecs : :class:`.Data`
        Matrix where each column is an initial Floquet mode.

    e_quasi : np.ndarray[float]
        The quasi energies of the Hamiltonian.
    """

    def __init__(
        self,
        H,
        T,
        args=None,
        options=None,
        sparse=False,
        sort=True,
        precompute=None,
    ):
        """
        Parameters
        ----------
        H : :obj:`.Qobj`, :obj:`.QobjEvo`, QobjEvo compatible format.
            System Hamiltonian, with period `T`.

        T : float
            Period of the Hamiltonian.

        args : None / *dictionary*
            dictionary of parameters for time-dependent Hamiltonians and
            collapse operators.

        options : dict [None]
            Options used by sesolve to compute the floquet modes.

        sparse : bool [False]
            Whether to use the sparse eigen solver when computing the
            quasi-energies.

        sort : bool [True]
            Whether to sort the quasi-energies.

        precompute : list [None]
            If provided, a list of time at which to store the propagators
            for later use when computing modes and states. Default is
            ``linspace(0, T, 101)`` corresponding to the default integration
            steps used for the floquet tensor computation.
        """
        if not T > 0:
            raise ValueError("The period need to be a positive number.")
        self.T = T
        if precompute is not None:
            tlist = np.unique(np.atleast_1d(precompute) % self.T)
            memoize = len(tlist)
            if tlist[0] != 0:
                memoize += 1
            if tlist[-1] != T:
                memoize += 1
        else:
            # Default computation
            tlist = np.linspace(0, T, 101)
            memoize = 101
        if (
            isinstance(H, QobjEvo)
            and (H._feedback_functions or H._solver_only_feedback)
        ):
            raise NotImplementedError("FloquetBasis does not support feedback")
        self.U = Propagator(H, args=args, options=options, memoize=memoize)
        for t in tlist:
            # Do the evolution by steps to save the intermediate results.
            self.U(t)
        U_T = self.U(self.T)
        if not sparse and isinstance(U_T.data, _data.CSR):
            U_T = U_T.to("Dense")
        evals, evecs = _data.eigs(U_T.data)
        e_quasi = -np.angle(evals) / T
        if sort:
            perm = np.argsort(e_quasi)
            self.evecs = _data.permute.indices(evecs, col_perm=np.argsort(perm))
            self.e_quasi = e_quasi[perm]
        else:
            self.evecs = evecs
            self.e_quasi = e_quasi

    def _as_ketlist(self, kets_mat):
        """
        Split the Data array in a list of kets.
        """
        dims = [self.U(0).dims[0], [1]]
        return [
            Qobj(ket, dims=dims)
            for ket in _data.split_columns(kets_mat)
        ]

    def mode(self, t, data=False):
        """
        Calculate the Floquet modes at time ``t``.

        Parameters
        ----------
        t : float
            The time for which to evaluate the Floquet mode.

        data : bool [False]
            Whether to return the states as a single data matrix or a list of
            ket states.

        Returns
        -------
        output : list[:obj:`.Qobj`], :class:`.Data`
            A list of Floquet states for the time ``t`` or the states as column
            in a single matrix.
        """
        t = t % self.T
        if t == 0.0:
            kets_mat = self.evecs
        else:
            U = self.U(t).data
            phases = _data.diag(np.exp(1j * t * self.e_quasi))
            kets_mat = U @ self.evecs @ phases
        if data:
            return kets_mat
        else:
            return self._as_ketlist(kets_mat)

    def state(self, t, data=False):
        """
        Evaluate the floquet states at time t.

        Parameters
        ----------
        t : float
            The time for which to evaluate the Floquet states.

        data : bool [False]
            Whether to return the states as a single data matrix or a list of
            ket states.

        Returns
        -------
        output : list[:obj:`.Qobj`], :class:`.Data`
            A list of Floquet states for the time ``t`` or the states as column
            in a single matrix.
        """
        if t:
            phases = _data.diag(np.exp(-1j * t * self.e_quasi))
            states_mat = self.mode(t, True) @ phases
        else:
            states_mat = self.evecs
        if data:
            return states_mat
        else:
            return self._as_ketlist(states_mat)

    def from_floquet_basis(self, floquet_basis, t=0):
        """
        Transform a ket or density matrix from the Floquet basis at time ``t``
        to the lab basis.

        Parameters
        ----------
        floquet_basis : :obj:`.Qobj`, :class:`.Data`
            Initial state in the Floquet basis at time ``t``. May be either a
            ket or density matrix.

        t : float [0]
            The time at which to evaluate the Floquet states.

        Returns
        -------
        output : :obj:`.Qobj`, :class:`.Data`
            The state in the lab basis. The return type is the same as the type
            of the input state.
        """
        is_Qobj = isinstance(floquet_basis, Qobj)
        if is_Qobj:
            dims = floquet_basis.dims
            floquet_basis = floquet_basis.data
            if dims[0] != self.U(0).dims[1]:
                raise ValueError(
                    "Dimensions of the state do not match the Hamiltonian"
                )

        state_mat = self.state(t, True)
        lab_basis = state_mat @ floquet_basis
        if floquet_basis.shape[1] != 1:
            lab_basis = lab_basis @ state_mat.adjoint()

        if is_Qobj:
            return Qobj(lab_basis, dims=dims)
        return lab_basis

    def to_floquet_basis(self, lab_basis, t=0):
        """
        Transform a ket or density matrix in the lab basis
        to the Floquet basis at time ``t``.

        Parameters
        ----------
        lab_basis : :obj:`.Qobj`, :class:`.Data`
            Initial state in the lab basis.

        t : float [0]
            The time at which to evaluate the Floquet states.

        Returns
        -------
        output : :obj:`.Qobj`, :class:`.Data`
            The state in the Floquet basis. The return type is the same as the
            type of the input state.
        """
        is_Qobj = isinstance(lab_basis, Qobj)
        if is_Qobj:
            dims = lab_basis.dims
            lab_basis = lab_basis.data
            if dims[0] != self.U(0).dims[1]:
                raise ValueError(
                    "Dimensions of the state do not match the Hamiltonian"
                )

        state_mat = self.state(t, True)
        floquet_basis = state_mat.adjoint() @ lab_basis
        if lab_basis.shape[1] != 1:
            floquet_basis = floquet_basis @ state_mat

        if is_Qobj:
            return Qobj(floquet_basis, dims=dims)
        return floquet_basis


def _floquet_delta_tensor(f_energies, kmax, T):
    """
    Floquet-Markov master equation X matrices.

    Parameters
    ----------
    f_energies : np.ndarray
        The Floquet energies.

    kmax : int
        The truncation of the number of sidebands (default 5).

    T : float
        The period of the time-dependence of the Hamiltonian.

    Returns
    -------
    delta : np.ndarray
        Floquet delta tensor.
    """
    delta = np.subtract.outer(f_energies, f_energies)
    return np.add.outer(delta, np.arange(-kmax, kmax + 1) * (2 * np.pi / T))


def _floquet_X_matrices(floquet_basis, c_ops, kmax, ntimes=100):
    """
    Floquet-Markov master equation X matrices.

    Parameters
    ----------
    floquet_basis : :class:`FloquetBasis`
        The system Hamiltonian wrapped in a FloquetBasis object.

    c_ops : list of :obj:`.Qobj`
        The collapse operators describing the dissipation.

    kmax : int
        The truncation of the number of sidebands (default 5).

    ntimes : int [100]
        The number of integration steps (for calculating X) within one period.

    Returns
    -------
    X : list of dict of :class:`.Data`
        A dict of the sidebands ``k`` for the X matrices of each c_ops
    """
    T = floquet_basis.T
    N = floquet_basis.U(0).shape[0]
    omega = (2 * np.pi) / T
    tlist = np.linspace(T / ntimes, T, ntimes)
    ks = np.arange(-kmax, kmax + 1)
    out = {k: [_data.csr.zeros(N, N)] * len(c_ops) for k in ks}

    for t in tlist:
        mode = floquet_basis.mode(t, data=True)
        FFs = [mode.adjoint() @ c_op.data @ mode for c_op in c_ops]
        for k, phi in zip(ks, np.exp(-1j * ks * omega * t) / ntimes):
            out[k] = [
                _data.add(prev, new, phi) for prev, new in zip(out[k], FFs)
            ]

    return [{k: out[k][i] for k in ks} for i in range(len(c_ops))]


def _floquet_gamma_matrices(X, delta, J_cb):
    """
    Floquet-Markov master equation gamma matrices.

    Parameters
    ----------
    X : list of dict of :class:`.Data`
        Floquet X matrices created by :func:`_floquet_X_matrices`.

    delta : np.ndarray
        Floquet delta tensor created by :func:`_floquet_delta_tensor`.

    J_cb : list of callables
        A list callback functions that compute the noise power spectrum as
        a function of frequency. The list should contain one callable for each
        collapse operator `c_op`, in the same order as the elements of `X`.
        Each callable should accept a numpy array of frequencies and return a
        numpy array of corresponding noise power.

    Returns
    -------
    gammas : dict of :class:`.Data`
        A dict mapping the sidebands ``k`` to their gamma matrices.
    """
    N = delta.shape[0]
    kmax = (delta.shape[2] - 1) // 2
    gamma = {k: _data.csr.zeros(N, N) for k in range(-kmax, kmax + 1, 1)}

    for X_c_op, sp in zip(X, J_cb):
        response = sp(delta) * ((2 + 0j) * np.pi)
        response = [
            _data.Dense(response[:, :, k], copy=False)
            for k in range(2 * kmax + 1)
        ]
        for k in range(-kmax, kmax + 1, 1):
            gamma[k] = _data.add(
                gamma[k],
                _data.multiply(
                    _data.multiply(X_c_op[k].conj(), X_c_op[k]),
                    response[k + kmax],
                ),
            )
    return gamma


def _floquet_A_matrix(delta, gamma, w_th):
    """
    Floquet-Markov master equation rate matrix.

    Parameters
    ----------
    delta : np.ndarray
        Floquet delta tensor created by :func:`_floquet_delta_tensor`.

    gamma : dict of :class:`.Data`
        Floquet gamma matrices created by :func:`_floquet_gamma_matrices`.

    w_th : float
        The temperature in units of frequency.
    """
    kmax = (delta.shape[2] - 1) // 2

    if w_th > 0.0:
        deltap = np.copy(delta)
        deltap[deltap == 0.0] = np.inf
        thermal = 1.0 / (np.exp(np.abs(deltap) / w_th) - 1.0)
        thermal = [_data.Dense(thermal[:, :, k]) for k in range(2 * kmax + 1)]

        gamma_kk = _data.add(gamma[0], gamma[0].transpose())
        A = _data.add(gamma[0], _data.multiply(thermal[kmax], gamma_kk))

        for k in range(1, kmax + 1):
            g_kk = _data.add(gamma[k], gamma[-k].transpose())
            thermal_kk = _data.multiply(thermal[kmax + k], g_kk)
            A = _data.add(A, _data.add(gamma[k], thermal_kk))
            thermal_kk = _data.multiply(thermal[kmax - k], g_kk.transpose())
            A = _data.add(A, _data.add(gamma[-k], thermal_kk))
    else:
        # w_th is 0, thermal = 0s
        A = gamma[0]
        for k in range(1, kmax + 1):
            A = _data.add(gamma[k], A)
            A = _data.add(gamma[-k], A)

    return A


def _floquet_master_equation_tensor(A):
    """
    Construct a tensor that represents the master equation in the floquet
    basis (with constant Hamiltonian and collapse operators?).

    Simplest RWA approximation [Grifoni et al, Phys.Rep. 304 229 (1998)]

    Parameters
    ----------
    A : :class:`.Data`
        Floquet-Markov master equation rate matrix.

    Returns
    -------
    output : array
        The Floquet-Markov master equation tensor `R`.
    """
    N = A.shape[0]

    # R[i+N*i, j+N*j] = A[j, i]
    cols = np.arange(N, dtype=np.int32)
    rows = np.linspace(1 - 1 / (N + 1), N, N**2 + 1, dtype=np.int32)
    data = np.ones(N, dtype=complex)
    expand = _data.csr.CSR((data, cols, rows), shape=(N**2, N))

    R = expand @ A.transpose() @ expand.transpose()

    # S[i+N*j, j+N*i] = -1/2 * sum_k(A[i, k] + A[j, k])
    ket_1 = _data.Dense(np.ones(N, dtype=complex))
    Asum = A @ ket_1
    to_super = _data.add(_data.kron(Asum, ket_1), _data.kron(ket_1, Asum))
    S = _data.diag(to_super.to_array().flatten() * -0.5, 0)

    return _data.add(R, S)


def floquet_tensor(H, c_ops, spectra_cb, T=0, w_th=0.0, kmax=5, nT=100):
    """
    Construct a tensor that represents the master equation in the floquet
    basis.

    Simplest RWA approximation [Grifoni et al, Phys.Rep. 304 229 (1998)]

    Parameters
    ----------
    H : :obj:`.QobjEvo`, :obj:`.FloquetBasis`
        Periodic Hamiltonian a floquet basis system.

    T : float, optional
        The period of the time-dependence of the hamiltonian. Optional if ``H``
        is a ``FloquetBasis`` object.

    c_ops : list of :class:`.Qobj`
        list of collapse operators.

    spectra_cb : list callback functions
        List of callback functions that compute the noise power spectrum as
        a function of frequency for the collapse operators in `c_ops`.

    w_th : float, default: 0.0
        The temperature in units of frequency.

    kmax : int, default: 5
        The truncation of the number of sidebands (default 5).

    nT : int, default: 100
        The number of integration steps (for calculating X) within one period.

    Returns
    -------
    output : array
        The Floquet-Markov master equation tensor `R`.
    """
    if isinstance(H, FloquetBasis):
        floquet_basis = H
        T = H.T
    else:
        floquet_basis = FloquetBasis(H, T)
    energy = floquet_basis.e_quasi
    delta = _floquet_delta_tensor(energy, kmax, T)
    x = _floquet_X_matrices(floquet_basis, c_ops, kmax, nT)
    gamma = _floquet_gamma_matrices(x, delta, spectra_cb)
    a = _floquet_A_matrix(delta, gamma, w_th)
    r = _floquet_master_equation_tensor(a)
    dims = floquet_basis.U(0).dims
    return Qobj(r, dims=[dims, dims], superrep="super", copy=False)


def fsesolve(H, psi0, tlist, e_ops=None, T=0.0, args=None, options=None):
    """
    Solve the Schrodinger equation using the Floquet formalism.

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        Periodic system Hamiltonian as :obj:`.QobjEvo`. List of
        [:obj:`.Qobj`, :obj:`.Coefficient`] or callable that
        can be made into :obj:`.QobjEvo` are also accepted.

    psi0 : :class:`.Qobj`
        Initial state vector (ket). If an operator is provided,

    tlist : *list* / *array*
        List of times for :math:`t`.

    e_ops : list of :class:`.Qobj` / callback function, optional
        List of operators for which to evaluate expectation values. If this
        list is empty, the state vectors for each time in `tlist` will be
        returned instead of expectation values.

    T : float, default=tlist[-1]
        The period of the time-dependence of the hamiltonian.

    args : dictionary, optional
        Dictionary with variables required to evaluate H.

    options : dict, optional
        Options for the results.

        - | store_final_state : bool
          | Whether or not to store the final state of the evolution in the
            result class.
        - | store_states : bool, None
          | Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.
        - | normalize_output : bool
          | Normalize output state to hide ODE numerical errors.

    Returns
    -------
    output : :class:`.Result`
        An instance of the class :class:`.Result`, which
        contains either an *array* of expectation values or an array of
        state vectors, for the times specified by `tlist`.
    """
    if isinstance(H, FloquetBasis):
        floquet_basis = H
    else:
        T = T or tlist[-1]
        # `fsesolve` is a fallback from `fmmesolve`, for the later, options
        # are for the open system evolution.
        floquet_basis = FloquetBasis(H, T, args, precompute=tlist)

    f_coeff = floquet_basis.to_floquet_basis(psi0)
    result_options = {
        "store_final_state": False,
        "store_states": None,
        "normalize_output": True,
    }
    result_options.update(options or {})
    result = Result(e_ops, result_options, solver="fsesolve")
    for t in tlist:
        state_t = floquet_basis.from_floquet_basis(f_coeff, t)
        result.add(t, state_t)

    return result


def fmmesolve(
    H,
    rho0,
    tlist,
    c_ops=None,
    e_ops=None,
    spectra_cb=None,
    T=0,
    w_th=0.0,
    args=None,
    options=None,
):
    """
    Solve the dynamics for the system using the Floquet-Markov master equation.

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        Periodic system Hamiltonian as :obj:`.QobjEvo`. List of
        [:obj:`.Qobj`, :obj:`.Coefficient`] or callable that
        can be made into :obj:`.QobjEvo` are also accepted.

    rho0 / psi0 : :class:`.Qobj`
        Initial density matrix or state vector (ket).

    tlist : *list* / *array*
        List of times for :math:`t`.

    c_ops : list of :class:`.Qobj`, optional
        List of collapse operators. Time dependent collapse operators are not
        supported. Fall back on :func:`fsesolve` if not provided.

    e_ops : list of :class:`.Qobj` / callback function, optional
        List of operators for which to evaluate expectation values.
        The states are reverted to the lab basis before applying the

    spectra_cb : list callback functions, default: ``lambda w: (w > 0)``
        List of callback functions that compute the noise power spectrum as
        a function of frequency for the collapse operators in `c_ops`.

    T : float, default=tlist[-1]
        The period of the time-dependence of the hamiltonian. The default value
        ``0`` indicates that the 'tlist' spans a single period of the driving.

    w_th : float, default: 0.0
        The temperature of the environment in units of frequency.
        For example, if the Hamiltonian written in units of 2pi GHz, and the
        temperature is given in K, use the following conversion:

            temperature = 25e-3 # unit K
            h = 6.626e-34
            kB = 1.38e-23
            args['w_th'] = temperature * (kB / h) * 2 * pi * 1e-9

    args : dict, optional
        Dictionary of parameters for time-dependent Hamiltonian

    options : dict, optional
        Dictionary of options for the solver.

        - | store_final_state : bool
          | Whether or not to store the final state of the evolution in the
            result class.
        - | store_states : bool, None
          | Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.
        - | store_floquet_states : bool
          | Whether or not to store the density matrices in the floquet basis
            in ``result.floquet_states``.
        - | normalize_output : bool
          | Normalize output state to hide ODE numerical errors.
        - | progress_bar : str {'text', 'enhanced', 'tqdm', ''}
          | How to present the solver progress.
            'tqdm' uses the python module of the same name and raise an error
            if not installed. Empty string or False will disable the bar.
        - | progress_kwargs : dict
          | kwargs to pass to the progress_bar. Qutip's bars use `chunk_size`.
        - | method : str ["adams", "bdf", "lsoda", "dop853", "vern9", etc.]
          | Which differential equation integration method to use.
        - | atol, rtol : float
          | Absolute and relative tolerance of the ODE integrator.
        - | nsteps : int
          | Maximum number of (internally defined) steps allowed in one
            ``tlist`` step.
        - | max_step : float
          | Maximum lenght of one internal step. When using pulses, it should
            be less than half the width of the thinnest pulse.

        Other options could be supported depending on the integration method,
        see `Integrator <./classes.html#classes-ode>`_.

    Returns
    -------
    result: :class:`.Result`

        An instance of the class :class:`.Result`, which contains
        the expectation values for the times specified by ``tlist``, and/or the
        state density matrices corresponding to the times.
    """
    if c_ops is None and rho0.isket:
        return fsesolve(
            H,
            rho0,
            tlist,
            e_ops=e_ops,
            T=T,
            args=args,
            options=options,
        )

    if isinstance(H, FloquetBasis):
        floquet_basis = H
    else:
        T = T or tlist[-1]
        t_precompute = np.concatenate([tlist, np.linspace(0, T, 101)])
        # `fsesolve` is a fallback from `fmmesolve`, for the later, options
        # are for the open system evolution.
        floquet_basis = FloquetBasis(H, T, args, precompute=t_precompute)

    if not w_th and args:
        w_th = args.get("w_th", 0.0)

    if isinstance(c_ops, Qobj):
        c_ops = [c_ops]

    if spectra_cb is None:
        spectra_cb = [lambda w: (w > 0)]
    elif callable(spectra_cb):
        spectra_cb = [spectra_cb]
    if len(spectra_cb) == 1:
        spectra_cb = spectra_cb * len(c_ops)

    a_ops = list(zip(c_ops, spectra_cb))

    solver = FMESolver(floquet_basis, a_ops, w_th=w_th, options=options)
    return solver.run(rho0, tlist, e_ops=e_ops)


class FloquetResult(Result):
    def _post_init(self, floquet_basis):
        self.floquet_basis = floquet_basis
        if self.options["store_floquet_states"]:
            self.floquet_states = []
        else:
            self.floquet_states = None
        super()._post_init()

    def add(self, t, state):
        if self.options["store_floquet_states"]:
            self.floquet_states.append(state)

        state = self.floquet_basis.from_floquet_basis(state, t)
        super().add(t, state)


class FMESolver(MESolver):
    """
    Solver for the Floquet-Markov master equation.

    .. note ::
        Operators (``c_ops`` and ``e_ops``) are in the laboratory basis.

    Parameters
    ----------
    floquet_basis : :class:`.FloquetBasis`
        The system Hamiltonian wrapped in a FloquetBasis object. Choosing a
        different integrator for the ``floquet_basis`` than for the evolution
        of the floquet state can improve the performance.

    a_ops : list of tuple(:class:`.Qobj`, callable)
        List of collapse operators and the corresponding function for the noise
        power spectrum. The collapse operator must be a :obj:`.Qobj` and
        cannot be time dependent. The spectrum function must take and return
        an numpy array.

    w_th : float
        The temperature of the environment in units of Hamiltonian frequency.

    kmax : int [5]
        The truncation of the number of sidebands..

    nT : int [20*kmax]
        The number of integration steps (for calculating X) within one period.

    options : dict, optional
        Options for the solver, see :obj:`FMESolver.options` and
        `Integrator <./classes.html#classes-ode>`_ for a list of all options.
    """

    name = "fmmesolve"
    _avail_integrators = {}
    _resultclass = FloquetResult
    solver_options = {
        "progress_bar": "text",
        "progress_kwargs": {"chunk_size": 10},
        "store_final_state": False,
        "store_states": None,
        "normalize_output": True,
        "method": "adams",
        "store_floquet_states": False,
    }

    def __init__(
        self, floquet_basis, a_ops, w_th=0.0, *, kmax=5, nT=None, options=None
    ):
        self.options = options
        if isinstance(floquet_basis, FloquetBasis):
            self.floquet_basis = floquet_basis
        else:
            raise TypeError("The ``floquet_basis`` must be a FloquetBasis")

        nT = nT or max(100, 20 * kmax)
        self._num_collapse = len(a_ops)
        c_ops, spectra_cb = zip(*a_ops)
        if not all(
            isinstance(c_op, Qobj) and callable(spectrum)
            for c_op, spectrum in a_ops
        ):
            raise TypeError("a_ops must be tuple of (Qobj, callable)")
        self.rhs = QobjEvo(
            floquet_tensor(
                self.floquet_basis,
                c_ops,
                spectra_cb,
                w_th=w_th,
                kmax=kmax,
                nT=nT,
            )
        )

        self._integrator = self._get_integrator()
        self._state_metadata = {}
        self.stats = self._initialize_stats()

    def _initialize_stats(self):
        stats = Solver._initialize_stats(self)
        stats.update(
            {
                "solver": "Floquet-Markov master equation",
                "num_collapse": self._num_collapse,
            }
        )
        return stats

    def _argument(self, args):
        if args:
            raise ValueError("FMESolver cannot update arguments")

    def start(self, state0, t0, *, floquet=False):
        """
        Set the initial state and time for a step evolution.
        ``options`` for the evolutions are read at this step.

        Parameters
        ----------
        state0 : :obj:`.Qobj`
            Initial state of the evolution.

        t0 : double
            Initial time of the evolution.

        floquet : bool, optional {False}
            Whether the initial state is in the floquet basis or laboratory
            basis.
        """
        if not floquet:
            state0 = self.floquet_basis.to_floquet_basis(state0, t0)
        super().start(state0, t0)

    def step(self, t, *, args=None, copy=True, floquet=False):
        """
        Evolve the state to ``t`` and return the state as a :obj:`.Qobj`.

        Parameters
        ----------
        t : double
            Time to evolve to, must be higher than the last call.

        copy : bool, optional {True}
            Whether to return a copy of the data or the data in the ODE solver.

        floquet : bool, optional {False}
            Whether to return the state in the floquet basis or laboratory
            basis.

        args : dict, optional {None}
            Not supported

        Notes
        -----
        The state must be initialized first by calling ``start`` or
        ``run``. If ``run`` is called, ``step`` will continue from the last
        time and state obtained.
        """
        if args:
            raise ValueError("FMESolver cannot update arguments")
        state = super().step(t)
        if not floquet:
            state = self.floquet_basis.from_floquet_basis(state, t)
        elif copy:
            state = state.copy()
        return state

    def run(self, state0, tlist, *, floquet=False, args=None, e_ops=None):
        """
        Calculate the evolution of the quantum system.

        For a ``state0`` at time ``tlist[0]`` do the evolution as directed by
        ``rhs`` and for each time in ``tlist`` store the state and/or
        expectation values in a :class:`.Result`. The evolution method and
        stored results are determined by ``options``.

        Parameters
        ----------
        state0 : :obj:`.Qobj`
            Initial state of the evolution.

        tlist : list of double
            Time for which to save the results (state and/or expect) of the
            evolution. The first element of the list is the initial time of the
            evolution. Each times of the list must be increasing, but does not
            need to be uniformy distributed.

        floquet : bool, optional {False}
            Whether the initial state in the floquet basis or laboratory basis.

        args : dict, optional {None}
            Not supported

        e_ops : list {None}
            List of Qobj, QobjEvo or callable to compute the expectation
            values. Function[s] must have the signature
            f(t : float, state : Qobj) -> expect.

        Returns
        -------
        results : :class:`FloquetResult`
            Results of the evolution. States and/or expect will be saved. You
            can control the saved data in the options.
        """
        if args:
            raise ValueError("FMESolver cannot update arguments")
        if not floquet:
            state0 = self.floquet_basis.to_floquet_basis(state0, tlist[0])
        _time_start = time()
        _data0 = self._prepare_state(state0)
        self._integrator.set_state(tlist[0], _data0)
        stats = self._initialize_stats()
        results = self._resultclass(
            e_ops,
            self.options,
            solver=self.name,
            stats=stats,
            floquet_basis=self.floquet_basis,
        )
        results.add(tlist[0], self._restore_state(_data0, copy=False))
        stats["preparation time"] += time() - _time_start

        progress_bar = progress_bars[self.options["progress_bar"]](
            len(tlist) - 1, **self.options["progress_kwargs"]
        )
        for t, state in self._integrator.run(tlist):
            progress_bar.update()
            results.add(t, self._restore_state(state, copy=False))
        progress_bar.finished()

        stats["run time"] = progress_bar.total_time()
        # TODO: It would be nice if integrator could give evolution statistics
        # stats.update(_integrator.stats)
        return results

    @classmethod
    def ExpectFeedback(cls):
        """
        Expect of the state of the evolution to be used in a time-dependent
        operator.

        Not not implemented for FMESolver
        """
        raise NotImplementedError(
            "Feedback not implemented for floquet solver."
        )

    @classmethod
    def StateFeedback(cls):
        """
        State of the evolution to be used in a time-dependent operator.

        Not not implemented for FMESolver
        """
        raise NotImplementedError(
            "Feedback not implemented for floquet solver."
        )



import numpy as np
import scipy.linalg
import scipy.sparse as sp
import scipy.sparse.linalg
from itertools import combinations

from .dense import Dense, from_csr
from .csr import CSR
from .properties import isherm as _isherm
from qutip.settings import settings

__all__ = [
    'eigs', 'eigs_csr', 'eigs_dense',
    'svd', 'svd_csr', 'svd_dense',
]


def _orthogonalize(vec, other):
    cross = np.sum(np.conj(other) * vec)
    vec -= cross * other
    norm = np.sum(np.conj(vec) * vec)**0.5
    vec /= norm


if settings.eigh_unsafe:
    def eigh(mat, eigvals=None):
        val, vec = scipy.linalg.eig(mat)
        val = np.real(val)
        idx = np.argsort(val)
        val = val[idx]
        vec = vec[:, idx]
        if eigvals:
            val = val[eigvals[0]:eigvals[1]+1]
            vec = vec[:, eigvals[0]:eigvals[1]+1]
        same_eigv = 0
        for i in range(1, len(val)):
            if abs(val[i] - val[i-1]) < 1e-12:
                same_eigv += 1
                for j in range(same_eigv):
                    _orthogonalize(vec[:, i], vec[:, i-j-1])
            else:
                same_eigv = 0
        return val, vec

    def eigvalsh(a, eigvals=None):
        val = scipy.linalg.eigvals(a)
        val = np.sort(np.real(val))
        if eigvals:
            return val[eigvals[0]:eigvals[1]+1]
        return val
else:
    eigh = scipy.linalg.eigh
    eigvalsh = scipy.linalg.eigvalsh


def _eigs_dense(data, isherm, vecs, eigvals, num_large, num_small):
    """
    Internal functions for computing eigenvalues and eigenstates for a dense
    matrix.
    """
    N = data.shape[0]
    kwargs = {}
    if eigvals != 0 and isherm:
        kwargs['subset_by_index'] = (
            [0, num_small-1] if num_small else [N-num_large, N-1]
        )
    if vecs:
        driver = eigh if isherm else scipy.linalg.eig
        evals, evecs = driver(data, **kwargs)
    else:
        driver = eigvalsh if isherm else scipy.linalg.eigvals
        evals = driver(data, **kwargs)
        evecs = None

    _zipped = list(zip(evals, range(len(evals))))
    _zipped.sort()
    evals, perm = list(zip(*_zipped))

    if vecs:
        evecs = np.array([evecs[:, k] for k in perm]).T

    if not isherm and eigvals > 0:
        if vecs:
            if num_small > 0:
                evals, evecs = evals[:num_small], evecs[:, :num_small]
            elif num_large > 0:
                evals = evals[(N - num_large):]
                evecs = evecs[:, (N - num_large):]
        else:
            if num_small > 0:
                evals = evals[:num_small]
            elif num_large > 0:
                evals = evals[(N - num_large):]
    return np.array(evals), evecs


def _eigs_csr(data, isherm, vecs, eigvals, num_large, num_small, tol, maxiter):
    """
    Internal functions for computing eigenvalues and eigenstates for a sparse
    matrix.
    """
    N = data.shape[0]
    big_vals = np.array([])
    small_vals = np.array([])
    evecs = None

    remove_one = 0  # 0: remove none, 1: remove smallest, -1: remove largest
    if eigvals == (N - 1):
        # calculate all eigenvalues and remove one at output if using sparse
        # 1: remove the smallest, -1, remove the largest
        remove_one = 1 if (num_small > 0) else -1
        eigvals = 0
        num_small = num_large = N // 2
        num_small += N % 2

    if vecs:
        if isherm:
            if num_large > 0:
                big_vals, big_vecs = sp.linalg.eigsh(data, k=num_large,
                                                     which='LA', tol=tol,
                                                     maxiter=maxiter)
            if num_small > 0:
                small_vals, small_vecs = sp.linalg.eigsh(
                    data, k=num_small, which='SA',
                    tol=tol, maxiter=maxiter)

        else:
            if num_large > 0:
                big_vals, big_vecs = sp.linalg.eigs(data, k=num_large,
                                                    which='LR', tol=tol,
                                                    maxiter=maxiter)
            if num_small > 0:
                small_vals, small_vecs = sp.linalg.eigs(
                    data, k=num_small, which='SR',
                    tol=tol, maxiter=maxiter)

        if num_large != 0 and num_small != 0:
            evecs = np.hstack([small_vecs, big_vecs])
        elif num_large != 0 and num_small == 0:
            evecs = big_vecs
        elif num_large == 0 and num_small != 0:
            evecs = small_vecs
    else:
        if isherm:
            if num_large > 0:
                big_vals = sp.linalg.eigsh(
                    data, k=num_large, which='LA',
                    return_eigenvectors=False, tol=tol, maxiter=maxiter)
            if num_small > 0:
                small_vals = sp.linalg.eigsh(
                    data, k=num_small, which='SA',
                    return_eigenvectors=False, tol=tol, maxiter=maxiter)
        else:
            if num_large > 0:
                big_vals = sp.linalg.eigs(
                    data, k=num_large, which='LR',
                    return_eigenvectors=False, tol=tol, maxiter=maxiter)
            if num_small > 0:
                small_vals = sp.linalg.eigs(
                    data, k=num_small, which='SR',
                    return_eigenvectors=False, tol=tol, maxiter=maxiter)

    evals = np.hstack((small_vals, big_vals))
    if isherm:
        evals = np.real(evals)

    _zipped = list(zip(evals, range(len(evals))))
    _zipped.sort()
    evals, perm = list(zip(*_zipped))

    if vecs:
        evecs = np.array([evecs[:, k] for k in perm]).T

    # remove last element if requesting N-1 eigs and using sparse
    if remove_one == 1:
        evals = evals[:-1]
        if vecs:
            evecs = evecs[:, :-1]
    elif remove_one == -1:
        evals = evals[1:]
        if vecs:
            evecs = evecs[:, 1:]

    return np.array(evals), evecs


def _eigs_check_shape(data):
    if data.shape[0] != data.shape[1]:
        raise TypeError("Can only diagonalize square matrices")


def _eigs_fix_eigvals(data, eigvals, sort):
    N = data.shape[0]
    if eigvals == N:
        eigvals = 0
    if eigvals > N:
        raise ValueError("Number of requested eigen vals/vecs must be <= N.")
    if sort not in ('low', 'high'):
        raise ValueError("'sort' must be 'low' or 'high'")

    # set number of large and small eigenvals/vecs
    if eigvals == 0:  # user wants all eigs (default)
        num_small = num_large = N // 2
        num_small += N % 2
    else:  # if user wants only a few eigen vals/vecs
        num_small, num_large = (eigvals, 0) if sort == 'low' else (0, eigvals)
    return eigvals, num_large, num_small


def eigs_csr(data, /, isherm=None, vecs=True, sort='low', eigvals=0,
             tol=0, maxiter=100000):
    """
    Return eigenvalues and eigenvectors for a CSR matrix.  This specialisation
    may take some extra keyword arguments in addition to the full documentation
    specified in :func:`.eigs`.

    This method is typically slower and less accurate than the dense eigenvalue
    solver; you probably want that, unless memory concerns deem it impossible.

    Extra keyword arguments
    -----------------------
    tol : float (0)
        Tolerance for sparse eigensolver.  Sufficiently small tolerances (such
        as 0) cause the solver to use machine precision.
    maxiter : int (100_000)
        Max number of iterations used by sparse eigensolver.
    """
    if not isinstance(data, CSR):
        raise TypeError("expected data in CSR format but got "
                        + str(type(data)))
    if data.shape[0] < 4:
        # For small matrix, the sparse solver can't compute all eigenvalues.
        return eigs_dense(from_csr(data), isherm, vecs, sort, eigvals)
    _eigs_check_shape(data)
    eigvals, num_large, num_small = _eigs_fix_eigvals(data, eigvals, sort)
    isherm = isherm if isherm is not None else _isherm(data)
    evals, evecs = _eigs_csr(data.as_scipy(), isherm, vecs, eigvals,
                             num_large, num_small, tol, maxiter)

    if vecs and isherm:
        i = 0
        while i < len(evals):
            num_degen = np.sum(np.abs(evals - evals[i]) < (2 * tol or 1e-14))
            # orthogonalize vectors 1 .. k with respect to the first, then
            # 2 .. k with respect to the second, and so on. Relies on both the
            # order of each pair and the ordering of pairs returned by
            # combinations.
            for k, l in combinations(range(num_degen), 2):
                _orthogonalize(evecs[:, i+l], evecs[:, i+k])
            i += num_degen

    if sort == 'high':
        # Flip arrays around.
        if vecs:
            evecs = np.fliplr(evecs)
        evals = evals[::-1]
    return (evals, Dense(evecs, copy=False)) if vecs else evals


def eigs_dense(data, /, isherm=None, vecs=True, sort='low', eigvals=0):
    """
    Return eigenvalues and eigenvectors for a Dense matrix.  Takes no special
    keyword arguments; see the primary documentation in :func:`.eigs`.
    """
    if not isinstance(data, Dense):
        raise TypeError("expected data in Dense format but got "
                        + str(type(data)))
    _eigs_check_shape(data)
    eigvals, num_large, num_small = _eigs_fix_eigvals(data, eigvals, sort)
    isherm = isherm if isherm is not None else _isherm(data)
    evals, evecs = _eigs_dense(data.as_ndarray(), isherm, vecs, eigvals,
                               num_large, num_small)
    if sort == 'high':
        # Flip arrays around.
        if vecs:
            evecs = np.fliplr(evecs)
        evals = evals[::-1]
    return (evals, Dense(evecs, copy=False)) if vecs else evals


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect


# We use eigs_dense as the signature source, since in this case it has the
# complete signature that we allow, so we don't need to manually set it.
eigs = _Dispatcher(eigs_dense, name='eigs', inputs=('data',), out=False)
eigs.__doc__ =\
    """
    Return eigenvalues and (optionally) eigenvectors for a data-layer object.

    Some particular specialisations of this function may take additional
    keyword arguments (such as the CSR solver).  See their particular
    docstrings for details on those.

    Parameters
    ----------
    data : Data
        Input matrix
    isherm : bool, optional
        Indicate whether the matrix is Hermitian or not.  There are special
        Hermitian eigenvalue and -vector solvers for this case, which will take
        care of orthonormalisation and ensuring better accuracy.  If this is
        not specified either way, it will be calculated from the data.
    vecs : bool, optional (True)
        Whether the eigenvectors should be returned as well.
    sort : {'low', 'high'}, optional
        Sort the output of the eigenvalues and -vectors ordered by the relevant
        size of the real part of the eigenvalue from 'low' to high or from
        'high' to low.  If not all of the eigenvalues are requested, this
        influences which eigenvalues will be found.
    eigvals : int, optional
        Number of eigenvalues and -vectors to return.  If `0`, then returns
        all.

    Returns
    -------
    eigenvalues : np.ndarray
        The requested eigenvalues, sorted in the expected order.  The dtype is
        `np.complex128`, unless `isherm=True`, in which case it will be
        `np.float64`.
    eigenvectors : Data
        Only if `vecs=True`.  An array of the eigenvectors corresponding to the
        order of the eigenvalues.
    """
eigs.add_specialisations([
    (CSR, eigs_csr),
    (Dense, eigs_dense),
], _defer=True)


def svd_csr(data, vecs=True, k=6, **kw):
    """
    Singular Value Decomposition:

    ``data = U @ S @ Vh``

    Where ``S`` is diagonal.

    Parameters
    ----------
    data : Data
        Input matrix
    vecs : bool, optional (True)
        Whether the singular vectors (``U``, ``Vh``) should be returned.
    k : int, optional (6)
        Number of state to compute, default is ``6`` to match scipy's default.
    **kw : dict
        Options to pass to ``scipy.sparse.linalg.svds``.

    Returns
    -------
    U : Dense
        Left singular vectors as columns. Only returned if ``vecs == True``.
        shape = (data.shape[0], k)
    S : np.ndarray
        The ``k``'s largest singular values.
    Vh : Dense
        Right singular vectors as rows. Only returned if ``vecs == True``.
        shape = (k, data.shape[1])

    .. note::
        svds cannot compute all states at once. While it could find the
        largest and smallest in 2 calls, it has issues converging with when
        solving for the smallest (finding the 5 smallest in a 50x50 matrix
        can fail with default options). It should be used when not all states
        are needed.
    """
    out = scipy.sparse.linalg.svds(
        data.as_scipy(), k, return_singular_vectors=vecs, **kw
    )
    if vecs:
        u, s, vh = out
        return Dense(u, copy=False), s, Dense(vh, copy=False)
    return out


def svd_dense(data, vecs=True, **kw):
    """
    Singular Value Decomposition:

    ``data = U @ S @ Vh``

    Where ``S`` is diagonal.

    Parameters
    ----------
    data : Data
        Input matrix
    vecs : bool, optional (True)
        Whether the singular vectors (``U``, ``Vh``) should be returned.
    **kw : dict
        Options to pass to ``scipy.linalg.svd``.

    Returns
    -------
    U : Dense
        Left singular vectors as columns. Only returned if ``vecs == True``.
    S : np.ndarray
        Singular values.
    Vh : Dense
        Right singular vectors as rows. Only returned if ``vecs == True``.
    """
    out = scipy.linalg.svd(
        data.to_array(), compute_uv=vecs, **kw
    )
    if vecs:
        u, s, vh = out
        return Dense(u, copy=False), s, Dense(vh, copy=False)
    return out


svd = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('data', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('vecs', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
    ]),
    name='svd',
    module=__name__,
    inputs=('data',),
    out=False)
svd.__doc__ =\
    """
    Singular Value Decomposition:

    ``data = U @ S @ Vh``

    Where ``S`` is diagonal.

    Parameters
    ----------
    data : Data
        Input matrix
    vecs : bool, optional (True)
        Whether the singular vectors (``U``, ``Vh``) should be returned.

    Returns
    -------
    U : Dense
        Left singular vectors as columns. Only returned if ``vecs == True``.
    S : np.ndarray
        Singular values.
    Vh : Dense
        Right singular vectors as rows. Only returned if ``vecs == True``.
    """
# Dense implementation return all states, but sparse implementation compute
# only a few states. So only the dense version is registered.
svd.add_specialisations([
    (Dense, svd_dense),
], _defer=True)


del _Dispatcher
del _inspect



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport cython
from cpython cimport mem

from qutip.settings import settings

from qutip.core.data.base cimport idxint
from qutip.core.data cimport csr, dense, CSR, Dense, Dia
from qutip.core.data.adjoint cimport transpose_csr

cdef extern from *:
    # Not defined in cpython.mem for some reason, but is in pymem.h.
    void *PyMem_Calloc(size_t nelem, size_t elsize)

__all__ = [
    'isherm', 'isherm_csr', 'isherm_dense', 'isherm_dia',
    'isdiag', 'isdiag_csr', 'isdiag_dense', 'isdiag_dia',
    'iszero', 'iszero_csr', 'iszero_dense', 'iszero_dia',
]

cdef inline bint _conj_feq(double complex a, double complex b, double tol) nogil:
    """Check whether a == conj(b) up to an absolute tolerance."""
    cdef double re = a.real - b.real
    cdef double im = a.imag + b.imag
    # Comparing the squares should be fine---there is possible precision loss
    # in the addition, but as re*re and im*im are both positive semidefinite,
    # the floating point result is strictly not greater than the true value.
    # Save the cycles: don't sqrt.
    return re*re + im*im < tol*tol

cdef inline bint _feq_zero(double complex a, double tol) nogil:
    return a.real*a.real + a.imag*a.imag < tol*tol

cdef inline double _abssq(double complex x) nogil:
    return x.real*x.real + x.imag*x.imag


cdef bint _isherm_csr_full(CSR matrix, double tol) except 2:
    """
    Full, structural test for Hermicity of a matrix.  We assume that the input
    matrix has already had its shape tested (must be square).

    This test is only necessary when there is at least one value which is less
    than the tolerance, and needed to be compared to an implicit zero.  In
    general it is less efficient than the other test, and allocates more
    memory.
    """
    cdef CSR transpose = transpose_csr(matrix)
    cdef idxint row, ptr_a, ptr_b, col_a, col_b
    for row in range(matrix.shape[0]):
        ptr_a, ptr_a_end = matrix.row_index[row], matrix.row_index[row + 1]
        ptr_b, ptr_b_end = transpose.row_index[row], transpose.row_index[row + 1]
        while ptr_a < ptr_a_end and ptr_b < ptr_b_end:
            # Doing this on every loop actually involves a few more
            # de-references than are strictly necessary, but just
            # simplifies the logic checking for the end of the row.
            col_a = matrix.col_index[ptr_a]
            col_b = transpose.col_index[ptr_b]
            if col_a == col_b:
                if not _conj_feq(matrix.data[ptr_a], transpose.data[ptr_b], tol):
                    return False
                ptr_a += 1
                ptr_b += 1
            elif col_a < col_b:
                if not _feq_zero(matrix.data[ptr_a], tol):
                    return False
                ptr_a += 1
            else:
                if not _feq_zero(transpose.data[ptr_b], tol):
                    return False
                ptr_b += 1
        for ptr_a in range(ptr_a, ptr_a_end):
            if not _feq_zero(matrix.data[ptr_a], tol):
                return False
        for ptr_b in range(ptr_b, ptr_b_end):
            if not _feq_zero(transpose.data[ptr_b], tol):
                return False
    return True


cpdef bint isherm_csr(CSR matrix, double tol=-1):
    """
    Determine whether an input CSR matrix is Hermitian up to a given
    floating-point tolerance.

    Parameters
    ----------
    matrix : CSR
        Input matrix to test
    tol : double, optional
        Absolute tolerance value to use.  Defaults to
        :obj:`settings.core['atol']`.

    Returns
    -------
    bint
        Boolean True if it is Hermitian, False if not.

    Notes
    -----
    The implementation is effectively just taking the adjoint, but rather than
    actually allocating and creating a new matrix, we just check whether the
    output would match the input matrix.  If we cannot be certain of Hermicity
    because the sizes of some elements are within tolerance of 0, we have to
    resort to a complete adjoint calculation.
    """
    tol = tol if tol >= 0 else settings.core["atol"]
    cdef size_t row, col, ptr, ptr_t, nrows=matrix.shape[0]
    if matrix.shape[0] != matrix.shape[1]:
        return False
    cdef idxint *out_row_index = <idxint *>PyMem_Calloc(nrows + 1, sizeof(idxint))
    if out_row_index == NULL:
        raise MemoryError
    matrix.sort_indices()
    try:
        for row in range(nrows):
            for ptr in range(matrix.row_index[row], matrix.row_index[row + 1]):
                col = matrix.col_index[ptr] + 1
                out_row_index[col] += 1
        for row in range(nrows):
            out_row_index[row+1] += out_row_index[row]
            if out_row_index[row + 1] != matrix.row_index[row + 1]:
                # Structures are not the same, but it could still be Hermitian
                # if any value is less than the tolerance.  That is the
                # worst-case scenario, so we sacrifice its speed in favour of
                # returning faster for the more common failure cases.
                for ptr in range(matrix.row_index[nrows]):
                    if _conj_feq(matrix.data[ptr], 0, tol):
                        return _isherm_csr_full(matrix, tol)
                return False
        for row in range(nrows):
            for ptr in range(matrix.row_index[row], matrix.row_index[row + 1]):
                col = matrix.col_index[ptr]
                # Pointer into the "transposed" matrix.
                ptr_t = out_row_index[col]
                out_row_index[col] += 1
                if row != matrix.col_index[ptr_t]:
                    return _isherm_csr_full(matrix, tol)
                if not _conj_feq(matrix.data[ptr], matrix.data[ptr_t], tol):
                    return False
        return True
    finally:
        mem.PyMem_Free(out_row_index)


cpdef bint isherm_dia(Dia matrix, double tol=-1) nogil:
    cdef double complex val, valT
    cdef size_t diag, other_diag, col, start, end, other_start
    if tol < 0:
        with gil:
            tol = settings.core["atol"]
    if matrix.shape[0] != matrix.shape[1]:
        return False
    for diag in range(matrix.num_diag):
        if matrix.offsets[diag] == 0:
            for col in range(matrix.shape[1]):
                val = valT = matrix.data[diag * matrix.shape[1] + col]
                if not _conj_feq(val, valT, tol):
                    return False
            continue
        other_diag = 0
        while other_diag < matrix.num_diag:
            if matrix.offsets[diag] == -matrix.offsets[other_diag]:
                break
            other_diag += 1

        if other_diag < diag:
            continue

        start = max(0, matrix.offsets[diag])
        end = min(matrix.shape[1], matrix.shape[0] + matrix.offsets[diag])

        if other_diag == matrix.num_diag:
            # No matching diag, should be 0
            for col in range(start, end):
                val = matrix.data[diag * matrix.shape[1] + col]
                if not _feq_zero(val, tol):
                    return False
            continue

        other_start = max(0, matrix.offsets[other_diag])
        for col in range(end - start):
            val = matrix.data[diag * matrix.shape[1] + col + start]
            valT = matrix.data[other_diag * matrix.shape[1] + col + other_start]
            if not _conj_feq(val, valT, tol):
                return False
    return True


cpdef bint isherm_dense(Dense matrix, double tol=-1):
    """
    Determine whether an input Dense matrix is Hermitian up to a given
    floating-point tolerance.

    Parameters
    ----------
    matrix : Dense
        Input matrix to test
    tol : double, optional
        Absolute tolerance value to use.  Defaults to
        :obj:`settings.core['atol']`.

    Returns
    -------
    bint
        Boolean True if it is Hermitian, False if not.
    """
    if matrix.shape[0] != matrix.shape[1]:
        return False
    tol = tol if tol >= 0 else settings.core["atol"]
    cdef size_t row, col, size=matrix.shape[0]
    for row in range(size):
        for col in range(row + 1):
            if not _conj_feq(
                matrix.data[col*size+row],
                matrix.data[row*size+col],
                tol
            ):
                return False
    return True


cpdef bint isdiag_dia(Dia matrix, double tol=-1) nogil:
    cdef size_t diag, start, end, col
    if tol < 0:
        with gil:
            tol = settings.core["atol"]
    cdef double tolsq = tol*tol
    for diag in range(matrix.num_diag):
        if matrix.offsets[diag] == 0:
            continue
        start = max(0, matrix.offsets[diag])
        end = min(matrix.shape[1], matrix.shape[0] + matrix.offsets[diag])
        for col in range(start, end):
            if _abssq(matrix.data[diag * matrix.shape[1] + col]) > tolsq:
                return False
    return True


cpdef bint isdiag_csr(CSR matrix) nogil:
    cdef size_t row, ptr_start, ptr_end=matrix.row_index[0]
    for row in range(matrix.shape[0]):
        ptr_start, ptr_end = ptr_end, matrix.row_index[row + 1]
        if ptr_end - ptr_start > 1:
            return False
        if ptr_end - ptr_start == 1:
            if matrix.col_index[ptr_start] != row:
                return False
    return True


cpdef bint isdiag_dense(Dense matrix) nogil:
    cdef size_t row, row_stride = 1 if matrix.fortran else matrix.shape[1]
    cdef size_t col, col_stride = matrix.shape[0] if matrix.fortran else 1
    for row in range(matrix.shape[0]):
        for col in range(matrix.shape[1]):
            if (col != row) and matrix.data[col * col_stride + row * row_stride] != 0.:
                return False
    return True


cpdef bint iszero_dia(Dia matrix, double tol=-1) nogil:
    cdef size_t diag, start, end, col
    if tol < 0:
        with gil:
            tol = settings.core["atol"]
    cdef double tolsq = tol*tol
    for diag in range(matrix.num_diag):
        start = max(0, matrix.offsets[diag])
        end = min(matrix.shape[1], matrix.shape[0] + matrix.offsets[diag])
        for col in range(start, end):
            if _abssq(matrix.data[diag * matrix.shape[1] + col]) > tolsq:
                return False
    return True


cpdef bint iszero_csr(CSR matrix, double tol=-1) nogil:
    cdef size_t ptr
    if tol < 0:
        with gil:
            tol = settings.core["atol"]
    tolsq = tol*tol
    for ptr in range(csr.nnz(matrix)):
        if _abssq(matrix.data[ptr]) > tolsq:
            return False
    return True


cpdef bint iszero_dense(Dense matrix, double tol=-1) nogil:
    cdef size_t ptr
    if tol < 0:
        with gil:
            tol = settings.core["atol"]
    tolsq = tol*tol
    for ptr in range(matrix.shape[0]*matrix.shape[1]):
        if _abssq(matrix.data[ptr]) > tolsq:
            return False
    return True


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

isherm = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('tol', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=-1),
    ]),
    name='isherm',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
isherm.__doc__ =\
    """
    Check if the matrix is Hermitian up to a optional element-wise absolute
    tolerance.  If the tolerance given is less than zero, the global settings
    value `qutip.settings.atol` will be used instead.

    Only square matrices can possibly be Hermitian.

    Parameters
    ----------
    matrix : Data
        The matrix to test for Hermicity.

    tol : real, optional
        If given, the absolute tolerance used to compare two values for
        equality.  If not given, or given and negative, the value of
        `qutip.settings.atol` is used instead.
    """
isherm.add_specialisations([
    (Dense, isherm_dense),
    (Dia, isherm_dia),
    (CSR, isherm_csr),
], _defer=True)

isdiag = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='isdiag',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
isdiag.__doc__ =\
    """
    Check if the matrix is diagonal.  The matrix need not be square to test.

    Parameters
    ----------
    matrix : Data
        The matrix to test for diagonality.
    """
isdiag.add_specialisations([
    (Dense, isdiag_dense),
    (Dia, isdiag_dia),
    (CSR, isdiag_csr),
], _defer=True)

iszero = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('tol', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=-1),
    ]),
    name='iszero',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
iszero.__doc__ =\
    """
    Test if this matrix is the zero matrix, up to a certain absolute tolerance.

    Parameters
    ----------
    matrix : Data
        The matrix to test.
    tol : real, optional
        The absolute tolerance to use when comparing to zero.  If not given, or
        less than 0, use the core setting `atol`.

    Returns
    -------
    bool
        Whether the matrix is equivalent to 0 under the given absolute
        tolerance.
    """
iszero.add_specialisations([
    (CSR, iszero_csr),
    (Dia, iszero_dia),
    (Dense, iszero_dense),
], _defer=True)

del _inspect, _Dispatcher



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

from qutip.core.data cimport CSR, Dense, Dia

cpdef CSR tidyup_csr(CSR matrix, double tol, bint inplace=*)
cpdef Dense tidyup_dense(Dense matrix, double tol, bint inplace=*)
cpdef Dia tidyup_dia(Dia matrix, double tol, bint inplace=*)



from . import Dense, CSR, Dia
from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect
try:
    from scipy.sparse import csr_array
except ImportError:
    csr_array = None
from scipy.sparse import csr_matrix


__all__ = ["extract"]


def extract_dense(matrix, format=None, copy=True):
    """
    Return an array representation of the Dense data object.

    Parameters
    ----------
    matrix : Data
        The matrix to convert to the given format.

    format : str {"ndarray"}, default="ndarray"
        Type of the output.

    copy : bool, default: True
        Whether to return a copy of the data. If False,
        a view of the data is returned when possible.
    """
    if format not in [None, "ndarray"]:
        raise ValueError(
            "Dense can only be extracted to 'ndarray'"
        )
    if copy:
        return matrix.to_array()
    else:
        return matrix.as_ndarray()


def extract_csr(matrix, format=None, copy=True):
    """
    Return the scipy's object ``csr_matrix``.

    Parameters
    ----------
    matrix : Data
        The matrix to convert to common type.

    format : str, {"csr_matrix"}
        Type of the output.

    copy : bool, default: True
        Whether to pass a copy of the object or not.
    """
    if format not in [None, "scipy_csr", "csr_matrix"]:
        raise ValueError(
            "CSR can only be extracted to 'csr_matrix'"
        )
    csr_mat = matrix.as_scipy()
    if copy:
        csr_mat = csr_mat.copy()
    return csr_mat


def extract_dia(matrix, format=None, copy=True):
    """
    Return the scipy's object ``dia_matrix``.

    Parameters
    ----------
    matrix : Data
        The matrix to convert to common type.

    format : str, {"dia_matrix"}
        Type of the output.

    copy : bool, default: True
        Whether to pass a copy of the object or not.
    """
    if format not in [None, "scipy_dia", "dia_matrix"]:
        raise ValueError(
            "Dia can only be extracted to 'dia_matrix'"
        )
    dia_mat = matrix.as_scipy()
    if copy:
        dia_mat = dia_mat.copy()
    return dia_mat


extract = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter(
            'format', _inspect.Parameter.POSITIONAL_OR_KEYWORD, default=None
        ),
        _inspect.Parameter(
            'copy', _inspect.Parameter.POSITIONAL_OR_KEYWORD, default=True
        )
    ]),
    name='extract',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
extract.__doc__ =\
    """
    Return the common representation of the data layer object: scipy's
    ``csr_matrix`` for ``CSR``, numpy array for ``Dense``, Jax's ``Array`` for
    ``JaxArray``, etc.

    Parameters
    ----------
    matrix : Data
        The matrix to convert to common type.

    format : str, default: None
        Type of the output, "ndarray" for ``Dense``, "csr_array" for ``CSR``.
        A ValueError will be raised if the format is not supported.

    copy : bool, default: True
        Whether to pass a copy of the object.
    """
extract.add_specialisations([
    (CSR, extract_csr),
    (Dia, extract_dia),
    (Dense, extract_dense),
], _defer=True)


del _Dispatcher, _inspect



import numpy as np
from numpy.random import SeedSequence, default_rng
import scipy.sparse as sp
import scipy.linalg as la
import pytest

from qutip import qeye, num, to_kraus, kraus_to_choi, CoreOptions, Qobj
from qutip import data as _data
from qutip.random_objects import (
    rand_herm,
    rand_unitary,
    rand_dm,
    rand_ket,
    rand_stochastic,
    rand_super,
    rand_super_bcsz,
    rand_kraus_map,
)


@pytest.fixture(params=[
    12,
    [8],
    [2, 2, 3],
    [[2], [2]]
], ids=["int", "list", "tensor", "super"])
def dimensions(request):
    return request.param


@pytest.fixture(
    params=list(_data.to.dtypes),
    ids=lambda dtype: str(dtype)[:-2].split(".")[-1]
)
def dtype(request):
    return request.param


def _assert_density(qobj, density):
    N = np.sum(qobj.full() != 0)
    density = N / np.prod(qobj.shape)
    assert density == pytest.approx(density, abs=0.2)


def _assert_metadata(random_qobj, dims, dtype=None, super=False, ket=False):
    if isinstance(dims, int):
        dims = [dims]
    N = np.prod(dims)
    if super and not isinstance(dims[0], list):
        target_dims_0 = [dims, dims]
        shape0 = N**2
    else:
        target_dims_0 = dims
        shape0 = N

    if ket:
        target_dims_1 = [1]
        shape1 = 1
    else:
        target_dims_1 = target_dims_0
        shape1 = shape0

    assert random_qobj.dims[0] == target_dims_0
    assert random_qobj.dims[1] == target_dims_1
    assert random_qobj.shape == (shape0, shape1)

    if dtype:
        assert isinstance(random_qobj.data, dtype)


@pytest.mark.repeat(3)
@pytest.mark.parametrize('density', [0.2, 0.8], ids=["sparse", "dense"])
@pytest.mark.parametrize('distribution', ["fill", "pos_def"])
def test_rand_herm(dimensions, density, distribution, dtype):
    """
    Random Qobjs: Hermitian matrix
    """
    random_qobj = rand_herm(
        dimensions,
        density=density,
        distribution=distribution,
        dtype=dtype
    )
    if distribution == "pos_def":
        assert all(random_qobj.eigenenergies() > -1e14)
    assert random_qobj.isherm
    assert _data.isherm(random_qobj.data)
    _assert_metadata(random_qobj, dimensions, dtype)
    _assert_density(random_qobj, density)


@pytest.mark.repeat(3)
@pytest.mark.parametrize('density', [0.2, 0.8], ids=["sparse", "dense"])
def test_rand_herm_Eigs(dimensions, density):
    """
    Random Qobjs: Hermitian matrix - Eigs given
    """
    N = np.prod(dimensions)
    eigs = np.random.random(N)
    eigs /= np.sum(eigs)
    eigs.sort()
    random_qobj = rand_herm(dimensions, density, "eigen", eigenvalues=eigs)
    np.testing.assert_allclose(random_qobj.eigenenergies(), eigs)
    # verify hermitian
    assert random_qobj.isherm
    assert _data.isherm(random_qobj.data)
    _assert_metadata(random_qobj, dimensions)
    _assert_density(random_qobj, density)


@pytest.mark.repeat(5)
@pytest.mark.parametrize('distribution', ["haar", "exp"])
@pytest.mark.parametrize('density', [0.2, 0.8])
def test_rand_unitary(dimensions, distribution, density, dtype):
    """
    Random Qobjs: Tests that unitaries are actually unitary.
    """
    N = np.prod(dimensions)
    random_qobj = rand_unitary(
        dimensions, distribution=distribution,
        density=density, dtype=dtype
    )
    I = qeye(dimensions)
    assert random_qobj * random_qobj.dag() == I
    _assert_metadata(random_qobj, dimensions, dtype)
    if distribution == "exp":
        _assert_density(random_qobj, density)


@pytest.mark.repeat(3)
@pytest.mark.parametrize(["distribution", "kw"], [
    pytest.param("ginibre", {"rank": 3}),
    pytest.param("ginibre", {"rank": 1}),
    pytest.param("hs", {"rank": 0}),
    pytest.param("pure", {"rank": 1, "density": 0.5}),
    pytest.param("eigen", {"eigenvalues": True}),
    pytest.param("herm", {"density": 0.7}),
    pytest.param("herm", {"density": 0.3}),
])
def test_rand_dm(dimensions, kw, dtype, distribution):
    """
    Random Qobjs: Density matrix
    """
    N = np.prod(dimensions)
    print(N, kw)
    if "eigenvalues" in kw:
        eigs = np.random.random(N)
        eigs /= np.sum(eigs)
        eigs.sort()
        kw["eigenvalues"] = eigs

    random_qobj = rand_dm(
        dimensions,
        distribution=distribution,
        dtype=dtype,
        **kw
    )
    assert abs(random_qobj.tr() - 1.0) < 1e-14
    # verify all eigvals are >=0
    assert all(random_qobj.eigenenergies() >= -1e-14)
    # verify hermitian
    assert random_qobj.isherm
    assert _data.isherm(random_qobj.data)

    _assert_metadata(random_qobj, dimensions, dtype)

    eigenvalues = random_qobj.eigenenergies()

    if "rank" in kw:
        print(kw["rank"], N, kw["rank"] or N)
        desired_rank = kw["rank"] or N  # 'hs' is full rank.
        rank = sum([abs(E) >= 1e-10 for E in eigenvalues])
        assert rank == desired_rank

    if "eigenvalues" in kw:
        np.testing.assert_allclose(eigenvalues, kw["eigenvalues"])

    if "density" in kw:
        _assert_density(random_qobj, kw["density"])


@pytest.mark.repeat(5)
@pytest.mark.parametrize('kind', ["left", "right"])
def test_rand_stochastic(dimensions, kind, dtype):
    """
    Random Qobjs: Test random stochastic
    """
    random_qobj = rand_stochastic(dimensions, kind=kind, dtype=dtype)
    axis = {"left":0, "right":1}[kind]
    np.testing.assert_allclose(
        np.sum(random_qobj.full(), axis=axis),
        1, atol=1e-14
    )
    _assert_metadata(random_qobj, dimensions, dtype)


@pytest.mark.repeat(5)
@pytest.mark.parametrize('distribution', ["haar", "fill"])
def test_rand_ket(dimensions, distribution, dtype):
    """
    Random Qobjs: Test random ket type and norm.
    """
    random_qobj = rand_ket(dimensions, distribution=distribution, dtype=dtype)

    target_type = "ket"
    if isinstance(dimensions, list) and isinstance(dimensions[0], list):
        target_type = "operator-ket"
    assert random_qobj.type == target_type
    assert abs(random_qobj.norm() - 1) < 1e-14

    if isinstance(dimensions, int):
        dims = [dimensions]
    N = np.prod(dimensions)
    _assert_metadata(random_qobj, dimensions, dtype, ket=True)


@pytest.mark.repeat(2)
@pytest.mark.parametrize('superrep', ["choi", "super"])
def test_rand_super(dimensions, dtype, superrep):
    """
    Random Qobjs: Super operator.
    """
    random_qobj = rand_super(dimensions, dtype=dtype, superrep=superrep)
    assert random_qobj.issuper
    with CoreOptions(atol=2e-9):
        assert random_qobj.iscptp
    assert random_qobj.superrep == superrep
    _assert_metadata(random_qobj, dimensions, dtype, super=True)


@pytest.mark.repeat(2)
@pytest.mark.parametrize('rank', [None, 4])
@pytest.mark.parametrize('superrep', ["choi", "super"])
def test_rand_super_bcsz(dimensions, dtype, rank, superrep):
    """
    Random Qobjs: Tests that BCSZ-random superoperators are CPTP.
    """

    random_qobj = rand_super_bcsz(dimensions, rank=rank,
                                  dtype=dtype, superrep=superrep)
    assert random_qobj.issuper
    with CoreOptions(atol=1e-9):
        assert random_qobj.iscptp
    assert random_qobj.superrep == superrep
    _assert_metadata(random_qobj, dimensions, dtype, super=True)
    if (
        not rank
        and isinstance(dimensions, list)
        and isinstance(dimensions[0], list)
    ):
        # dimensions = [[a], [a]], qobj.dims = [[[a], [a]], [[a], [a]]]
        rank = np.prod(dimensions)
    elif not rank:
        # dimensions = [a], qobj.dims = [[[a], [a]], [[a], [a]]]
        rank = np.prod(dimensions)**2
    rank = rank or N
    obtained_rank = len(to_kraus(random_qobj, tol=1e-13))
    assert obtained_rank == rank


@pytest.mark.parametrize("function", [
    pytest.param(rand_herm, id="rand_herm"),
    pytest.param(rand_unitary, id="rand_unitary"),
    pytest.param(rand_dm, id="rand_dm"),
    pytest.param(rand_ket, id="rand_ket"),
    pytest.param(rand_stochastic, id="rand_stochastic"),
    pytest.param(rand_super, id="rand_super"),
    pytest.param(rand_super_bcsz, id="rand_super_bcsz"),
])
@pytest.mark.parametrize("seed", [
    pytest.param(lambda : 123, id="int"),
    pytest.param(lambda : SeedSequence(123), id="SeedSequence"),
    pytest.param(lambda : default_rng(123), id="Generator")
])
def test_random_seeds(function, seed):
    """
    Random Qobjs: Random number generator seed
    """
    U0 = function(5, seed=seed())
    U1 = function(5, seed=None)
    U2 = function(5, seed=seed())
    assert U0 != U1
    assert U0 == U2


def test_kraus_map(dimensions, dtype):
    if isinstance(dimensions, list) and isinstance(dimensions[0], list):
        # Each element of a kraus map cannot be a super operators
        with pytest.raises(TypeError) as err:
            kmap = rand_kraus_map(dimensions, dtype=dtype)
        assert "super operator" in str(err.value)
    else:
        kmap = rand_kraus_map(dimensions, dtype=dtype)
        _assert_metadata(kmap[0], dimensions, dtype)
        with CoreOptions(atol=1e-9):
            assert kraus_to_choi(kmap).iscptp


dtype_names = list(_data.to._str2type.keys()) + list(_data.to.dtypes)
dtype_types = list(_data.to._str2type.values()) + list(_data.to.dtypes)
@pytest.mark.parametrize(['alias', 'dtype'], zip(dtype_names, dtype_types),
                         ids=[str(dtype) for dtype in dtype_names])
@pytest.mark.parametrize('func', [
    rand_herm,
    rand_unitary,
    rand_dm,
    rand_ket,
    rand_stochastic,
    rand_super,
    rand_super_bcsz,
    rand_kraus_map,
])
def test_random_dtype(func, alias, dtype):
    with CoreOptions(default_dtype=alias):
        object = func(2)
        if isinstance(object, Qobj):
            assert isinstance(object.data, dtype)
        else:
            for obj in object:
                assert isinstance(obj.data, dtype)



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

# The exported function `expect(op, state)` is equivalent to
# `inner_op(state.adjoint(), op, state)` if `state` is a ket, or
# `trace(op @ state)` if state is a density matrix, but it's optimised to not
# make unnecessary extra allocations, or calculate extra factors.

from libc.math cimport sqrt

cdef extern from "<complex>" namespace "std" nogil:
    double complex conj(double complex x)

from qutip.core.data.base cimport idxint, Data
from qutip.core.data cimport csr, CSR, Dense, Dia
from .inner import inner
from .trace import trace, trace_oper_ket
from .matmul import matmul

__all__ = [
    'expect', 'expect_csr', 'expect_dense', 'expect_dia', 'expect_data',
    'expect_csr_dense', 'expect_dia_dense',
    'expect_super', 'expect_super_csr', 'expect_super_dia', 'expect_super_dense',
    'expect_super_csr_dense', 'expect_super_dia_dense', 'expect_super_data',
]

cdef int _check_shape_ket(Data op, Data state) except -1 nogil:
    if (
        op.shape[1] != state.shape[0]  # Matrix multiplication
        or state.shape[1] != 1  # State is ket
        or op.shape[0] != op.shape[1]  # op must be square matrix
    ):
        raise ValueError("incorrect input shapes "
                         + str(op.shape) + " and " + str(state.shape))
    return 0

cdef int _check_shape_dm(Data op, Data state) except -1 nogil:
    if (
        op.shape[1] != state.shape[0]  # Matrix multiplication
        or state.shape[0] != state.shape[1]  # State is square
        or op.shape[0] != op.shape[1]  # Op is square
    ):
        raise ValueError("incorrect input shapes "
                         + str(op.shape) + " and " + str(state.shape))
    return 0

cdef int _check_shape_super(Data op, Data state) except -1 nogil:
    if state.shape[1] != 1:
        raise ValueError("expected a column-stacked matrix")
    if (
        op.shape[1] != state.shape[0]  # Matrix multiplication
        or op.shape[0] != op.shape[1]  # Square matrix
    ):
        raise ValueError("incompatible shapes "
                         + str(op.shape) + ", " + str(state.shape))
    return 0


cdef double complex _expect_csr_ket(CSR op, CSR state) except * nogil:
    """
    Perform the operation
        state.adjoint() @ op @ state
    for a ket `state` and a square operator `op`.
    """
    _check_shape_ket(op, state)
    cdef double complex out=0, sum=0, mul
    cdef size_t row, col, ptr_op, ptr_ket
    for row in range(state.shape[0]):
        ptr_ket = state.row_index[row]
        if ptr_ket == state.row_index[row + 1]:
            continue
        sum = 0
        mul = conj(state.data[ptr_ket])
        for ptr_op in range(op.row_index[row], op.row_index[row + 1]):
            col = op.col_index[ptr_op]
            ptr_ket = state.row_index[col]
            if ptr_ket != state.row_index[col + 1]:
                sum += op.data[ptr_op] * state.data[ptr_ket]
        out += mul * sum
    return out

cdef double complex _expect_csr_dm(CSR op, CSR state) except * nogil:
    """
    Perform the operation
        tr(op @ state)
    for an operator `op` and a density matrix `state`.
    """
    _check_shape_dm(op, state)
    cdef double complex out=0
    cdef size_t row, col, ptr_op, ptr_state
    for row in range(op.shape[0]):
        for ptr_op in range(op.row_index[row], op.row_index[row + 1]):
            col = op.col_index[ptr_op]
            for ptr_state in range(state.row_index[col], state.row_index[col + 1]):
                if state.col_index[ptr_state] == row:
                    out += op.data[ptr_op] * state.data[ptr_state]
                    break
    return out


cpdef double complex expect_super_csr(CSR op, CSR state) except *:
    """
    Perform the operation `tr(op @ state)` where `op` is supplied as a
    superoperator, and `state` is a column-stacked operator.
    """
    _check_shape_super(op, state)
    cdef double complex out = 0.0
    cdef size_t row=0, ptr, col
    cdef size_t n = <size_t> sqrt(state.shape[0])
    for _ in range(n):
        for ptr in range(op.row_index[row], op.row_index[row + 1]):
            col = op.col_index[ptr]
            if state.row_index[col] != state.row_index[col + 1]:
                out += op.data[ptr] * state.data[state.row_index[col]]
        row += n + 1
    return out


cpdef double complex expect_csr(CSR op, CSR state) except *:
    """
    Get the expectation value of the operator `op` over the state `state`.  The
    state can be either a ket or a density matrix.

    The expectation of a state is defined as the operation:
        state.adjoint() @ op @ state
    and of a density matrix:
        tr(op @ state)
    """
    if state.shape[1] == 1:
        return _expect_csr_ket(op, state)
    return _expect_csr_dm(op, state)

cdef double complex _expect_csr_dense_ket(CSR op, Dense state) except * nogil:
    _check_shape_ket(op, state)
    cdef double complex out=0, sum
    cdef size_t row, ptr
    for row in range(op.shape[0]):
        if op.row_index[row] == op.row_index[row + 1]:
            continue
        sum = 0
        for ptr in range(op.row_index[row], op.row_index[row + 1]):
            sum += op.data[ptr] * state.data[op.col_index[ptr]]
        out += sum * conj(state.data[row])
    return out

cdef double complex _expect_csr_dense_dm(CSR op, Dense state) except * nogil:
    _check_shape_dm(op, state)
    cdef double complex out=0
    cdef size_t row, ptr_op, ptr_state=0, row_stride, col_stride
    row_stride = 1 if state.fortran else state.shape[1]
    col_stride = state.shape[0] if state.fortran else 1
    for row in range(op.shape[0]):
        if op.row_index[row] == op.row_index[row + 1]:
            continue
        ptr_state = row * col_stride
        for ptr_op in range(op.row_index[row], op.row_index[row + 1]):
            out += op.data[ptr_op] * state.data[ptr_state + row_stride*op.col_index[ptr_op]]
    return out


cdef double complex _expect_dense_ket(Dense op, Dense state) except * nogil:
    _check_shape_ket(op, state)
    cdef double complex out=0, sum
    cdef size_t row, col, op_row_stride, op_col_stride
    op_row_stride = 1 if op.fortran else op.shape[1]
    op_col_stride = op.shape[0] if op.fortran else 1

    for row in range(op.shape[0]):
        sum = 0
        for col in range(op.shape[0]):
            sum += (op.data[row * op_row_stride + col * op_col_stride] *
                    state.data[col])
        out += sum * conj(state.data[row])
    return out

cdef double complex _expect_dense_dense_dm(Dense op, Dense state) except * nogil:
    _check_shape_dm(op, state)
    cdef double complex out=0
    cdef size_t row, col, op_row_stride, op_col_stride
    cdef size_t state_row_stride, state_col_stride
    state_row_stride = 1 if state.fortran else state.shape[1]
    state_col_stride = state.shape[0] if state.fortran else 1
    op_row_stride = 1 if op.fortran else op.shape[1]
    op_col_stride = op.shape[0] if op.fortran else 1

    for row in range(op.shape[0]):
        for col in range(op.shape[1]):
            out += op.data[row * op_row_stride + col * op_col_stride] * \
                   state.data[col * state_row_stride + row * state_col_stride]
    return out


cpdef double complex expect_csr_dense(CSR op, Dense state) except *:
    """
    Get the expectation value of the operator `op` over the state `state`.  The
    state can be either a ket or a density matrix.

    The expectation of a state is defined as the operation:
        state.adjoint() @ op @ state
    and of a density matrix:
        tr(op @ state)
    """
    if state.shape[1] == 1:
        return _expect_csr_dense_ket(op, state)
    return _expect_csr_dense_dm(op, state)


cpdef double complex expect_dense(Dense op, Dense state) except *:
    """
    Get the expectation value of the operator `op` over the state `state`.  The
    state can be either a ket or a density matrix.

    The expectation of a state is defined as the operation:
        state.adjoint() @ op @ state
    and of a density matrix:
        tr(op @ state)
    """
    if state.shape[1] == 1:
        return _expect_dense_ket(op, state)
    return _expect_dense_dense_dm(op, state)


cpdef double complex expect_super_csr_dense(CSR op, Dense state) except * nogil:
    """
    Perform the operation `tr(op @ state)` where `op` is supplied as a
    superoperator, and `state` is a column-stacked operator.
    """
    _check_shape_super(op, state)
    cdef double complex out=0
    cdef size_t row=0, ptr
    cdef size_t n = <size_t> sqrt(state.shape[0])
    for _ in range(n):
        for ptr in range(op.row_index[row], op.row_index[row + 1]):
            out += op.data[ptr] * state.data[op.col_index[ptr]]
        row += n + 1
    return out


cpdef double complex expect_super_dense(Dense op, Dense state) except * nogil:
    """
    Perform the operation `tr(op @ state)` where `op` is supplied as a
    superoperator, and `state` is a column-stacked operator.
    """
    _check_shape_super(op, state)
    cdef double complex out=0
    cdef size_t row=0, col, N = state.shape[0]
    cdef size_t n = <size_t> sqrt(state.shape[0])
    cdef size_t op_row_stride, op_col_stride
    op_row_stride = 1 if op.fortran else op.shape[1]
    op_col_stride = op.shape[0] if op.fortran else 1

    for _ in range(n):
        for col in range(N):
            out += op.data[row * op_row_stride + col * op_col_stride] * \
                   state.data[col]
        row += n + 1
    return out


cpdef double complex expect_dia(Dia op, Dia state) except *:
    cdef double complex expect = 0.
    cdef idxint diag_bra, diag_op, diag_ket, i, length
    cdef idxint start_op, start_state, end_op, end_state
    if state.shape[1] == 1:
        _check_shape_ket(op, state)
        # Since the ket is sparse and possibly unsorted. Taking the n'th
        # element of the state require a loop on the diags. Thus 3 loops are
        # needed.
        for diag_ket in range(state.num_diag):
          #if -state.offsets[diag_ket] >= op.shape[1]:
          #  continue
          for diag_bra in range(state.num_diag):
            for diag_op in range(op.num_diag):
              if state.offsets[diag_ket] - state.offsets[diag_bra] + op.offsets[diag_op] == 0:
                expect += (
                  conj(state.data[diag_bra * state.shape[1]])
                  * state.data[diag_ket * state.shape[1]]
                  * op.data[diag_op * op.shape[1] - state.offsets[diag_ket]]
                )
    else:
      _check_shape_dm(op, state)
      for diag_op in range(op.num_diag):
        for diag_state in range(state.num_diag):
          if op.offsets[diag_op] == -state.offsets[diag_state]:

            start_op = max(0, op.offsets[diag_op])
            start_state = max(0, state.offsets[diag_state])
            end_op = min(op.shape[1], op.shape[0] + op.offsets[diag_op])
            end_state = min(state.shape[1], state.shape[0] + state.offsets[diag_state])
            length = min(end_op - start_op, end_state - start_state)

            for i in range(length):
              expect += (
                op.data[diag_op * op.shape[1] + i + start_op]
                * state.data[diag_state * state.shape[1] + i + start_state]
              )
    return expect


cpdef double complex expect_dia_dense(Dia op, Dense state) except *:
    cdef double complex expect = 0.
    cdef idxint i, diag_op, start_op, end_op, strideR, stride, start_state
    if state.shape[1] == 1:
        _check_shape_ket(op, state)
        for diag_op in range(op.num_diag):
            start_op = max(0, op.offsets[diag_op])
            end_op = min(op.shape[1], op.shape[0] + op.offsets[diag_op])
            for i in range(start_op, end_op):
                expect += (
                  op.data[diag_op * op.shape[1] + i]
                  * state.data[i]
                  * conj(state.data[i - op.offsets[diag_op]])
                )
    else:
      _check_shape_dm(op, state)
      stride = state.shape[0] + 1
      strideR = state.shape[0] if state.fortran else 1
      for diag_op in range(op.num_diag):
        start_op = max(0, op.offsets[diag_op])
        end_op = min(op.shape[1], op.shape[0] + op.offsets[diag_op])
        start_state = -op.offsets[diag_op] * strideR
        for i in range(start_op, end_op):
          expect += (
            op.data[diag_op * op.shape[1] + i]
            * state.data[start_state + i * stride]
          )
    return expect


cpdef double complex expect_super_dia(Dia op, Dia state) except *:
    cdef double complex expect = 0.
    _check_shape_super(op, state)
    cdef idxint diag_op, diag_state
    cdef idxint stride = <size_t> sqrt(state.shape[0]) + 1
    for diag_op in range(op.num_diag):
      for diag_state in range(state.num_diag):
        if (
            -state.offsets[diag_state] < op.shape[1]
            and -op.offsets[diag_op] - state.offsets[diag_state] >= 0
            and (-op.offsets[diag_op] - state.offsets[diag_state]) % stride == 0
        ):
            expect += state.data[diag_state * state.shape[1]] * op.data[diag_op * op.shape[1] - state.offsets[diag_state]]

    return expect


cpdef double complex expect_super_dia_dense(Dia op, Dense state) except *:
    cdef double complex expect = 0.
    _check_shape_super(op, state)
    cdef idxint col, diag_op, start, end
    cdef idxint stride = <size_t> sqrt(state.shape[0]) + 1
    for diag_op in range(op.num_diag):
      start = max(0, op.offsets[diag_op])
      end = min(op.shape[1], op.shape[0] + op.offsets[diag_op])
      col = (((start - op.offsets[diag_op] - 1) // stride) + 1) * stride + op.offsets[diag_op]
      while col < end:
        expect += op.data[diag_op * op.shape[1] + col] * state.data[col]
        col += stride
    return expect


def expect_data(Data op, Data state):
    """
    Get the expectation value of the operator `op` over the state `state`.  The
    state can be either a ket or a density matrix.

    The expectation of a state is defined as the operation:
        state.adjoint() @ op @ state
    and of a density matrix:
        tr(op @ state)
    """
    if state.shape[1] == 1:
        _check_shape_ket(op, state)
        return inner(state, matmul(op, state))
    _check_shape_dm(op, state)
    return trace(matmul(op, state))


def expect_super_data(Data op, Data state):
    """
    Perform the operation `tr(op @ state)` where `op` is supplied as a
    superoperator, and `state` is a column-stacked operator.
    """
    _check_shape_super(op, state)
    return trace_oper_ket(matmul(op, state))


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

expect = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('op', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('state', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='expect',
    module=__name__,
    inputs=('op', 'state'),
    out=False,
)
expect.__doc__ =\
    """
    Get the expectation value of the operator `op` over the state `state`.  The
    state can be either a ket or a density matrix.  Returns a complex number.

    The expectation of a state is defined as the operation:
        state.adjoint() @ op @ state
    and of a density matrix:
        tr(op @ state)
    """
expect.add_specialisations([
    (CSR, CSR, expect_csr),
    (CSR, Dense, expect_csr_dense),
    (Dense, Dense, expect_dense),
    (Dia, Dense, expect_dia_dense),
    (Dia, Dia, expect_dia),
    (Data, Data, expect_data),
], _defer=True)

expect_super = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('op', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('state', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='expect_super',
    module=__name__,
    inputs=('op', 'state'),
    out=False,
)
expect_super.__doc__ =\
    """
    Perform the operation `tr(op @ state)` where `op` is supplied as a
    superoperator, and `state` is a column-stacked operator.  Returns a complex
    number.
    """
expect_super.add_specialisations([
    (CSR, CSR, expect_super_csr),
    (CSR, Dense, expect_super_csr_dense),
    (Dense, Dense, expect_super_dense),
    (Dia, Dense, expect_super_dia_dense),
    (Dia, Dia, expect_super_dia),
    (Data, Data, expect_super_data),
], _defer=True)

del _inspect, _Dispatcher


cdef double complex expect_data_dense(Data op, Dense state) except *:
    cdef double complex out
    if type(op) is CSR:
        out = expect_csr_dense(op, state)
    elif type(op) is Dense:
        out = expect_dense(op, state)
    else:
        out = expect(op, state)
    return out


cdef double complex expect_super_data_dense(Data op, Dense state) except *:
    cdef double complex out
    if type(op) is CSR:
        out = expect_super_csr_dense(op, state)
    elif type(op) is Dense:
        out = expect_super_dense(op, state)
    else:
        out = expect_super(op, state)
    return out



#cython: language_level=3

from qutip.core.data.csr cimport CSR

cpdef double complex inner_csr(CSR left, CSR right, bint scalar_is_ket=*) except *
cpdef double complex inner_op_csr(CSR left, CSR op, CSR right, bint scalar_is_ket=*) except *



import pytest
import qutip
import numpy as np
from scipy.special import sph_harm

mpl = pytest.importorskip("matplotlib")
plt = pytest.importorskip("matplotlib.pyplot")

def test_result_state():
    H = qutip.rand_dm(2)
    tlist = np.linspace(0, 3*np.pi, 2)
    results = qutip.mesolve(H, H, tlist, [], [])

    fig, ani = qutip.anim_fock_distribution(results)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_result_state_ValueError():
    H = qutip.rand_dm(2)
    tlist = np.linspace(0, 3*np.pi, 2)
    results = qutip.mesolve(H, H, tlist, [], [],
                            options={"store_states": False})

    text = 'Nothing to visualize. You might have forgotten ' +\
           'to set options={"store_states": True}.'
    with pytest.raises(ValueError) as exc_info:
        fig, ani = qutip.anim_fock_distribution(results)
    assert str(exc_info.value) == text


def test_anim_wigner_sphere():
    psi = qutip.rand_ket(5)
    wigner = qutip.wigner_transform(psi, 2, False, 50, ["x"])

    fig, ani = qutip.anim_wigner_sphere([wigner]*2)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_hinton():
    rho = qutip.rand_dm(5)
    rhos = [rho]*2

    fig, ani = qutip.anim_hinton(rhos)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_sphereplot():
    theta = np.linspace(0, np.pi, 90)
    phi = np.linspace(0, 2 * np.pi, 60)
    phi_mesh, theta_mesh = np.meshgrid(phi, theta)
    values = sph_harm(-1, 2, phi_mesh, theta_mesh).T
    fig, ani = qutip.anim_sphereplot([values]*2, theta, phi)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_matrix_histogram():
    rho = qutip.rand_dm(5)
    rhos = [rho]*2

    fig, ani = qutip.anim_matrix_histogram(rhos)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_fock_distribution():
    rho = qutip.rand_dm(5)
    rhos = [rho]*2

    fig, ani = qutip.anim_fock_distribution(rhos)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_wigner():
    rho = qutip.rand_dm(5)
    rhos = [rho]*2

    fig, ani = qutip.anim_wigner(rhos)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


@pytest.mark.filterwarnings(
    "ignore:The input coordinates to pcolor:UserWarning"
)
def test_anim_spin_distribution():
    j = 5
    psi = qutip.spin_state(j, -j)
    psi = qutip.spin_coherent(j, np.random.rand() * np.pi,
                              np.random.rand() * 2 * np.pi)
    theta = np.linspace(0, np.pi, 50)
    phi = np.linspace(0, 2 * np.pi, 50)
    Q, THETA, PHI = qutip.spin_q_function(psi, theta, phi)

    fig, ani = qutip.anim_spin_distribution([Q]*2, THETA, PHI)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_qubism():
    state = qutip.ket("01")

    fig, ani = qutip.anim_qubism([state]*2)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)


def test_anim_schmidt():
    state = qutip.ket("01")

    fig, ani = qutip.anim_schmidt([state]*2)
    plt.close()

    assert isinstance(fig, mpl.figure.Figure)
    assert isinstance(ani, mpl.animation.ArtistAnimation)



import pytest
import itertools
import qutip
from qutip.core.data.dispatch import Dispatcher, _constructed_specialisation
import qutip.core.data as _data


class pseudo_dipatched:
    def __init__(self, types, output):
        if output:
            self.output = types[-1]
            self.inputs = types[:-1]
        else:
            self.output = False
            self.inputs = types

    def __call__(self, *args, **kwargs):
        assert len(args) == len(self.inputs)
        assert not kwargs
        for got, expected in zip(args, self.inputs):
            assert isinstance(got, expected)

        if not self.output:
            return
        elif self.output is _data.Data:
            return _data.zeros(1,1)
        else:
            return _data.zeros[self.output](1,1)


def _test_name(arg):
    if isinstance(arg, bool):
        return str(arg)
    return ", ".join(spec.__name__ for spec in arg)


@pytest.mark.parametrize(['specialisation', "output"], [
    ((_data.Dense,), True),
    ((_data.Data,), True),
    ((_data.Dense,), False),
    ((_data.Data,), False),
    ((_data.Dense, _data.Dense), True),
    ((_data.Dense, _data.CSR), True),
    ((_data.Data, _data.Data), True),
    ((_data.Data, _data.Dense), True),
    ((_data.Dense, _data.Data), True),
    ((_data.Dense, _data.Dense), False),
    ((_data.Dense, _data.CSR), False),
    ((_data.Data, _data.Data), False),
    ((_data.Data, _data.Dense), False),
    ((_data.Dense, _data.Data), False),
    ((_data.Dense, _data.Dense, _data.Dense), True),
    ((_data.Dense, _data.CSR, _data.CSR), True),
    ((_data.Data, _data.Data, _data.Data), True),
    ((_data.Data, _data.Dense, _data.Dense), True),
    ((_data.Dense, _data.Data, _data.Data), True),
], ids=_test_name)
def test_build_full(specialisation, output):
    """
    Test that the dispatched function can always be called with any input type.
    """
    def f(a=None, b=None, c=None, /):
        """
        Doc
        """


    n_input = len(specialisation) - output
    dispatched = Dispatcher(f, ("a", "b", "c")[:n_input], output)
    dispatched.add_specialisations(
        [specialisation + (pseudo_dipatched(specialisation, output),)]
    )

    for in_types in itertools.product(_data.to.dtypes, repeat=n_input):
        ins = tuple(_data.zeros[dtype](1, 1) for dtype in in_types)

        if not output:
            out = dispatched(*ins)
            out = dispatched[in_types](*ins)

        else:
            out = dispatched(*ins)
            assert out is not None

            out = dispatched[in_types](*ins)
            assert out is not None

            if output:
                for out_dtype in _data.to.dtypes:

                    out = dispatched[in_types + (out_dtype,)](*ins)
                    if output:
                        assert isinstance(out, out_dtype)

                    out = dispatched(*ins, dtype=out_dtype)
                    if output:
                        assert isinstance(out, out_dtype)


def test_Data_low_priority_one_dispatch():
    class func():
        __name__ = "dummy name"
        def __call__(self, a, /):
            return _data.zeros[_data.Dense](1, 1)

    f_dense = func()
    f_data = func()

    dispatched = Dispatcher(f_dense, ("a",), False)
    dispatched.add_specialisations([
        (_data.Dense, f_dense), (_data.Data, f_data)]
    )

    assert dispatched[_data.Dense] is f_dense
    assert dispatched[_data.CSR] is f_data

    dispatched = Dispatcher(f_dense, (), True)
    dispatched.add_specialisations([
        (_data.Dense, f_dense), (_data.Data, f_data)]
    )

    assert dispatched[_data.Dense] is f_dense
    assert isinstance(dispatched[_data.CSR], _constructed_specialisation)


def test_Data_low_priority_two_dispatch():
    class func():
        __name__ = ""
        def __init__(self):
            self.count = 0

        def __call__(self, a=None, b=None, /):
            self.count += 1
            return _data.zeros[_data.Dense](1, 1)

    f_dense = func()
    f_mixed = func()
    f_data = func()

    dispatched = Dispatcher(f_dense, ("a", "b"), False)
    dispatched.add_specialisations([
        (_data.Dense, _data.Dense, f_dense),
        (_data.Dense, _data.Data, f_mixed),
        (_data.Data, _data.Data, f_data),
    ])

    assert dispatched[_data.Dense, _data.Dense] is f_dense
    assert dispatched[_data.Dense, _data.CSR] is f_mixed
    assert dispatched[_data.CSR, _data.Dense] is f_data
    assert dispatched[_data.CSR, _data.CSR] is f_data

    dispatched = Dispatcher(f_dense, ("a",), True)
    dispatched.add_specialisations([
        (_data.Dense, _data.Dense, f_dense),
        (_data.Dense, _data.Data, f_mixed),
        (_data.Data, _data.Data, f_data),
    ])

    assert dispatched[_data.Dense, _data.Dense] is f_dense
    assert dispatched[_data.CSR] is f_data
    assert dispatched[_data.Dense] is f_dense
    assert isinstance(
        dispatched[_data.Dense, _data.CSR], _constructed_specialisation
    )

    assert f_mixed.count == 0
    dispatched[_data.Dense, _data.CSR](_data.zeros[_data.Dense](1, 1))
    assert f_mixed.count == 1

    assert isinstance(
        dispatched[_data.CSR, _data.Dense], _constructed_specialisation
    )
    assert isinstance(
        dispatched[_data.CSR, _data.CSR], _constructed_specialisation
    )

    assert f_data.count == 0
    dispatched[_data.CSR, _data.Dense](_data.zeros[_data.CSR](1, 1))
    dispatched[_data.CSR, _data.CSR](_data.zeros[_data.CSR](1, 1))
    assert f_data.count == 1



# -*- coding: utf-8 -*-

# @author: Arne L. Grimsmo
# @email1: arne.grimsmo@gmail.com
# @organization: University of Sherbrooke

"""
This module is an implementation of the method introduced in [1], for
solving open quantum systems subject to coherent feedback with a single
discrete time-delay. This method is referred to as the ``memory cascade''
method in qutip.

[1] Arne L. Grimsmo, Phys. Rev. Lett 115, 060402 (2015)
"""

import numpy as np
import warnings

from qutip import (
    sprepost,
    Qobj,
    spre,
    spost,
    liouvillian,
    qeye,
    mesolve,
    propagator,
    composite,
    isket,
    ket2dm,
    tensor_contract,
)


__all__ = ["MemoryCascade"]


class MemoryCascade:
    """Class for running memory cascade simulations of open quantum systems
    with time-delayed coherent feedback.

    Attributes
    ----------
    H_S : :class:`.Qobj`
        System Hamiltonian (can also be a Liouvillian)

    L1 : :class:`.Qobj` / list of :class:`.Qobj`
        System operators coupling into the feedback loop. Can be a single
        operator or a list of operators.

    L2 : :class:`.Qobj` / list of :class:`.Qobj`
        System operators coupling out of the feedback loop. Can be a single
        operator or a list of operators. L2 must have the same length as L1.

    S_matrix: *array*
        S matrix describing which operators in L1 are coupled to which
        operators in L2 by the feedback channel. Defaults to an n by n identity
        matrix where n is the number of elements in L1/L2.

    c_ops_markov : :class:`.Qobj` / list of :class:`.Qobj`
        Decay operators describing conventional Markovian decay channels.
        Can be a single operator or a list of operators.

    integrator : str {'propagator', 'mesolve'}
        Integrator method to use. Defaults to 'propagator' which tends to be
        faster for long times (i.e., large Hilbert space).

    options : dict
        Generic solver options.
    """

    def __init__(
        self,
        H_S,
        L1,
        L2,
        S_matrix=None,
        c_ops_markov=None,
        integrator="propagator",
        options=None,
    ):

        if options is None:
            self.options = {"progress_bar": False}
        else:
            self.options = options

        self.H_S = H_S
        self.sysdims = H_S.dims
        if isinstance(L1, Qobj):
            self.L1 = [L1]
        else:
            self.L1 = L1
        if isinstance(L2, Qobj):
            self.L2 = [L2]
        else:
            self.L2 = L2
        if not len(self.L1) == len(self.L2):
            raise ValueError("L1 and L2 has to be of equal length.")
        if isinstance(c_ops_markov, Qobj):
            self.c_ops_markov = [c_ops_markov]
        else:
            self.c_ops_markov = c_ops_markov

        if S_matrix is None:
            self.S_matrix = np.identity(len(self.L1))
        else:
            self.S_matrix = S_matrix
        # create system identity superoperator
        self.Id = qeye(H_S.shape[0])
        self.Id.dims = self.sysdims
        self.Id = sprepost(self.Id, self.Id)
        self.store_states = self.options.get("store_states", False)
        self.integrator = integrator
        self._generators = {}

    def generator(self, k):
        if k not in self._generators:
            self._generators[k] = _generator(
                k, self.H_S, self.L1, self.L2, self.S_matrix, self.c_ops_markov
            )
        return self._generators[k]

    def propagator(self, t, tau, notrace=False):
        """
        Compute propagator for time t and time-delay tau

        Parameters
        ----------
        t : *float*
            current time

        tau : *float*
            time-delay

        notrace : *bool* {False}
            If this optional is set to True, a propagator is returned for a
            cascade of k systems, where :math:`(k-1) tau < t < k tau`.
            If set to False (default), a generalized partial trace is performed
            and a propagator for a single system is returned.
        Returns
        -------
        : :class:`.Qobj`
            time-propagator for reduced system dynamics
        """
        k = int(t / tau) + 1
        s = t - (k - 1) * tau
        G1 = self.generator(k)
        E0 = qeye(G1.dims[0])
        E = _integrate(
            G1, E0, 0.0, s, integrator=self.integrator, opt=self.options
        )
        if k > 1:
            G2 = self.generator(k - 1)
            G2 = composite(G2, self.Id)
            E = _integrate(
                G2, E, s, tau, integrator=self.integrator, opt=self.options
            )

        if not notrace:
            E = _genptrace(E, k)
        return E

    def outfieldpropagator(
        self, blist, tlist, tau, c1=None, c2=None, notrace=False
    ):
        r"""
        Compute propagator for computing output field expectation values
        <O_n(tn)...O_2(t2)O_1(t1)> for times t1,t2,... and
        O_i = I, b_out, b_out^\dagger, b_loop, b_loop^\dagger

        Parameters
        ----------
        blist : array_like
            List of integers specifying the field operators:
            0: I (nothing)
            1: b_out
            2: b_out^\dagger
            3: b_loop
            4: b_loop^\dagger

        tlist : array_like
            list of corresponding times t1,..,tn at which to evaluate the field
            operators

        tau : float
            time-delay

        c1 : :class:`.Qobj`
            system collapse operator that couples to the in-loop field in
            question (only needs to be specified if self.L1 has more than one
            element)

        c2 : :class:`.Qobj`
            system collapse operator that couples to the output field in
            question (only needs to be specified if self.L2 has more than one
            element)

        notrace : bool {False}
            If this optional is set to True, a propagator is returned for a
            cascade of k systems, where :math:`(k-1) tau < t < k tau`.
            If set to False (default), a generalized partial trace is performed
            and a propagator for a single system is returned.

        Returns
        -------
        : :class:`.Qobj`
            time-propagator for computing field correlation function
        """
        if c1 is None and len(self.L1) == 1:
            c1 = self.L1[0]
        else:
            raise ValueError(
                "Argument c1 has to be specified when more than"
                + "one collapse operator couples to the feedback"
                + "loop."
            )
        if c2 is None and len(self.L2) == 1:
            c2 = self.L2[0]
        else:
            raise ValueError(
                "Argument c1 has to be specified when more than"
                + "one collapse operator couples to the feedback"
                + "loop."
            )
        klist = []
        slist = []
        for t in tlist:
            klist.append(int(t / tau) + 1)
            slist.append(t - (klist[-1] - 1) * tau)
        kmax = max(klist)
        zipped = sorted(zip(slist, klist, blist))
        slist = [s for (s, k, b) in zipped]
        klist = [k for (s, k, b) in zipped]
        blist = [b for (s, k, b) in zipped]

        G1 = self.generator(kmax)
        sprev = 0.0
        E0 = qeye(G1.dims[0])
        E = E0
        for i, s in enumerate(slist):
            E = _integrate(
                G1, E, sprev, s, integrator=self.integrator, opt=self.options
            )
            l2 = _localop(c2, klist[i], kmax)
            if klist[i] == 1:
                l1 = l2 * 0.0
            else:
                l1 = _localop(c1, klist[i] - 1, kmax)
            if blist[i] == 0:
                superop = self.Id
            elif blist[i] == 1:
                superop = spre(l1 + l2)
            elif blist[i] == 2:
                superop = spost(l1.dag() + l2.dag())
            elif blist[i] == 3:
                superop = spre(l1)
            elif blist[i] == 4:
                superop = spost(l1.dag())
            else:
                raise ValueError(
                    "Allowed values in blist are 0, 1, 2, 3 and 4."
                )

            E = superop @ E
            sprev = s
        E = _integrate(
            G1, E, slist[-1], tau, integrator=self.integrator, opt=self.options
        )

        if not notrace:
            E = _genptrace(E, kmax)
        return E

    def rhot(self, rho0, t, tau):
        """
        Compute the reduced system density matrix :math:`\\rho(t)`

        Parameters
        ----------
        rho0 : :class:`.Qobj`
            initial density matrix or state vector (ket)

        t : float
            current time

        tau : float
            time-delay

        Returns
        -------
        : :class:`.Qobj`
            density matrix at time :math:`t`
        """
        if isket(rho0):
            rho0 = ket2dm(rho0)

        E = self.propagator(t, tau)
        return E(rho0)

    def outfieldcorr(self, rho0, blist, tlist, tau, c1=None, c2=None):
        r"""
        Compute output field expectation value
        <O_n(tn)...O_2(t2)O_1(t1)> for times t1,t2,... and
        O_i = I, b_out, b_out^\dagger, b_loop, b_loop^\dagger


        Parameters
        ----------
        rho0 : :class:`.Qobj`
            initial density matrix or state vector (ket).

        blist : array_like
            List of integers specifying the field operators:
            0: I (nothing)
            1: b_out
            2: b_out^\dagger
            3: b_loop
            4: b_loop^\dagger

        tlist : array_like
            list of corresponding times t1,..,tn at which to evaluate the field
            operators

        tau : float
            time-delay

        c1 : :class:`.Qobj`
            system collapse operator that couples to the in-loop field in
            question (only needs to be specified if self.L1 has more than one
            element)

        c2 : :class:`.Qobj`
            system collapse operator that couples to the output field in
            question (only needs to be specified if self.L2 has more than one
            element)

        Returns
        -------
        : complex
            expectation value of field correlation function
        """
        if isket(rho0):
            rho0 = ket2dm(rho0)

        E = self.outfieldpropagator(blist, tlist, tau)
        return (E(rho0)).tr()


def _localop(op, l, k):
    """
    Create a local operator on the l'th system by tensoring
    with identity operators on all the other k-1 systems
    """
    if l < 1 or l > k:
        raise IndexError("index l out of range")
    out = op
    if l > 1:
        out = qeye(op.dims[0] * (l - 1)) & out
    if l < k:
        out = out & qeye(op.dims[0] * (k - l))

    return out


def _genptrace(E, k):
    """
    Perform a gneralized partial trace on a superoperator E, tracing out all
    subsystems but one.
    """
    for l in range(k - 1):
        nsys = len(E.dims[0][0])
        E = tensor_contract(E, (0, 2 * nsys + 1), (nsys, 3 * nsys + 1))
    return E


def _generator(k, H, L1, L2, S=None, c_ops_markov=None):
    """
    Create a Liouvillian for a cascaded chain of k system copies
    """
    id = qeye(H.dims[0][0])
    Id = sprepost(id, id)
    if S is None:
        S = np.identity(len(L1))
    # create Lindbladian

    # first system
    L = liouvillian(None, [_localop(c, 1, k) for c in L2])
    for l in range(1, k):
        # Bare Hamiltonian
        Hl = _localop(H, l, k)
        L += liouvillian(Hl, [])
        # Markovian Decay channels
        if c_ops_markov is not None:
            for c in c_ops_markov:
                cl = _localop(c, l, k)
                L += liouvillian(None, [cl])
        # Cascade coupling
        c1 = np.array([_localop(c, l, k) for c in L1])
        c2 = np.array([_localop(c, l + 1, k) for c in L2])
        c2dag = np.array([c.dag() for c in c2])
        Hcasc = -0.5j * np.dot(c2dag, np.dot(S, c1))
        Hcasc += Hcasc.dag()
        Lvec = c2 + np.dot(S, c1)
        L += liouvillian(Hcasc, [c for c in Lvec])
    # last system
    L += liouvillian(_localop(H, k, k), [_localop(c, k, k) for c in L1])
    if c_ops_markov is not None:
        for c in c_ops_markov:
            cl = _localop(c, k, k)
            L += liouvillian(None, [cl])
    # return generator
    return L


def _integrate(L, E0, ti, tf, integrator="propagator", opt=None):
    """
    Basic ode integrator
    """
    opt = opt or {}
    if tf > ti:
        if integrator == "mesolve":
            opt["store_final_state"] = True
            sol = mesolve(L, E0, [ti, tf], options=opt)
            return sol.final_state
        elif integrator == "propagator":
            return propagator(L, (tf - ti), options=opt) @ E0
        else:
            raise ValueError(
                'integrator keyword must be either "propagator" or "mesolve"'
            )
    else:
        return E0



import numpy as np
import scipy.sparse
import pytest

from qutip.core import data, qeye, CoreOptions
from qutip.core.data import dia, Dense, Dia

from . import conftest

# We only choose a small subset of dtypes to test so it isn't crazy.
_dtype_complex = ['complex128']
_dtype_float = ['float64']
_dtype_int = ['int32', 'int64']
_dtype_uint = ['uint32']


# Set up some fixtures for automatic parametrisation.

@pytest.fixture(params=[
    pytest.param((1, 5), id='ket'),
    pytest.param((5, 1), id='bra'),
    pytest.param((5, 5), id='square'),
    pytest.param((2, 4), id='wide'),
    pytest.param((4, 2), id='tall'),
])
def shape(request): return request.param

@pytest.fixture(params=[
    pytest.param(0, id='empty'),
    pytest.param(1, id='full'),
    pytest.param(0.2, id='sparse'),
])
def density(request): return request.param


@pytest.fixture(scope='function')
def scipy_dia(shape, density):
    return conftest.random_scipy_dia(shape, density)


def _valid_scipy():
    """Arbitrary valid scipy Dia"""
    return conftest.random_scipy_dia((10, 10), 0.5)


def _valid_arg():
    """
    Arbitrary valid 3-tuple which is a valid `arg` parameter for __init__.
    """
    sci = _valid_scipy()
    return (sci.data, sci.offsets)


@pytest.fixture(scope='function')
def data_diag(shape, density):
    return conftest.random_diag(shape, density)


class TestClassMethods:
    def test_init_from_scipy(self, scipy_dia):
        """Test that __init__ can accept a scipy dia matrix."""
        out = dia.Dia(scipy_dia)
        assert out.shape == scipy_dia.shape
        assert (out.as_scipy() - scipy_dia).nnz == 0

    def test_init_from_tuple(self, scipy_dia):
        """
        Test that __init__ does not throw when passed a 3-tuple.  Also tests
        the as_scipy() method succeeds.
        """
        arg = (scipy_dia.data, scipy_dia.offsets)
        out = dia.Dia(arg, shape=scipy_dia.shape)
        assert out.shape == scipy_dia.shape
        assert (out.as_scipy() - scipy_dia).nnz == 0

    @pytest.mark.parametrize('d_type', (
        _dtype_complex + _dtype_float + _dtype_int + _dtype_uint
    ))
    @pytest.mark.parametrize('o_type', _dtype_int + _dtype_uint)
    def test_init_from_tuple_allowed_dtypes(self, d_type, o_type):
        """
        Test that initialisation can use a variety of dtypes and converts into
        the correct type.
        """
        sci = _valid_scipy()
        data = sci.data.real.astype(d_type, casting='unsafe')
        offsets = sci.offsets.astype(o_type, casting='unsafe')
        scipy_dia = scipy.sparse.dia_matrix((data, offsets), shape=sci.shape)
        out = dia.Dia((data, offsets), shape=sci.shape)
        out_scipy = out.as_scipy()
        assert out.shape == scipy_dia.shape
        assert out_scipy.data.dtype == np.complex128
        assert (out_scipy - scipy_dia).nnz == 0

    @pytest.mark.parametrize(['arg', 'kwargs', 'error'], [
        pytest.param((), {}, ValueError, id="arg 0 tuple"),
        pytest.param((None,), {}, ValueError, id="arg 1 tuple"),
        pytest.param((None,)*2, {}, TypeError, id="arg None tuple"),
        pytest.param((None,)*3, {}, ValueError, id="arg 3 tuple"),
        pytest.param(_valid_scipy(), {'shape': ()}, ValueError,
                     id="scipy-shape 0 tuple"),
        pytest.param(_valid_scipy(), {'shape': (1,)}, ValueError,
                     id="scipy-shape 1 tuple"),
        pytest.param(_valid_scipy(), {'shape': (None, None)}, ValueError,
                     id="scipy-shape None tuple"),
        pytest.param(_valid_scipy(), {'shape': [2, 2]}, ValueError,
                     id="scipy-shape list"),
        pytest.param(_valid_scipy(), {'shape': (1, 2, 3)}, ValueError,
                     id="scipy-shape 3 tuple"),
        pytest.param(_valid_arg(), {'shape': ()}, ValueError,
                     id="arg-shape 0 tuple"),
        pytest.param(_valid_arg(), {'shape': (1,)}, ValueError,
                     id="arg-shape 1 tuple"),
        pytest.param(_valid_arg(), {'shape': (None, None)}, ValueError,
                     id="arg-shape None tuple"),
        pytest.param(_valid_arg(), {'shape': [2, 2]}, TypeError,
                     id="arg-shape list"),
        pytest.param(_valid_arg(), {'shape': (1, 2, 3)}, ValueError,
                     id="arg-shape 3 tuple"),
        pytest.param(_valid_arg(), {'shape': (-1, -1)}, ValueError,
                     id="arg-negative shape"),
    ])
    def test_init_from_wrong_input(self, arg, kwargs, error):
        """
        Test that the __init__ method raises a suitable error when passed
        incorrectly formatted inputs.

        This test also serves as a *partial* check that Dia safely handles
        deallocation in the presence of exceptions in its __init__ method.  If
        the tests segfault, it's quite likely that the memory management isn't
        being done correctly in the hand-off us setting our data buffers up and
        marking the numpy actually owns the data.
        """
        with pytest.raises(error):
            dia.Dia(arg, **kwargs)

    def test_copy_returns_a_correct_copy(self, data_diag):
        """
        Test that the copy() method produces an actual copy, and that the
        result represents the same matrix.
        """
        original = data_diag
        copy = data_diag.copy()
        assert original is not copy
        assert np.all(original.to_array() == copy.to_array())

    def test_as_scipy_returns_a_view(self, data_diag):
        """
        Test that modifying the views in the result of as_scipy() also modifies
        the underlying data structures.  This is important for allowing minor
        data modifications from within Python-space.
        """
        unmodified_copy = data_diag.copy()
        data_diag.as_scipy().data += 1
        modified_copy = data_diag.copy()
        assert np.any(data_diag.to_array() != unmodified_copy.to_array())
        assert np.all(data_diag.to_array() == modified_copy.to_array())

    def test_as_scipy_caches_result(self, data_diag):
        """
        Test that the as_scipy() method always returns the same view, even if
        called multiple times.
        """
        assert data_diag.as_scipy() is data_diag.as_scipy()

    def test_as_scipy_of_dia_from_scipy_is_different(self, scipy_dia):
        """
        Test that we produce a new scipy matrix, regardless of how we have
        initialised the type.
        """
        assert dia.Dia(scipy_dia).as_scipy() is not scipy_dia

    def test_as_scipy_of_copy_is_different(self, data_diag):
        """
        Test that as_scipy() does not return the same array, or the same views
        if it's not the same input matrix.  We don't want two Dia matrices to
        be linked.
        """
        original = data_diag.as_scipy()
        copy = data_diag.copy().as_scipy()
        assert original is not copy
        assert not np.may_share_memory(original.data, copy.data)
        assert not np.may_share_memory(original.offsets, copy.offsets)

    def test_as_scipy_is_correct_result(self, scipy_dia):
        """
        Test that as_scipy is actually giving the matrix we expect for a given
        input.
        """
        data_diag = dia.Dia(scipy_dia)
        assert isinstance(data_diag.as_scipy(), scipy.sparse.dia_matrix)
        assert (data_diag.as_scipy() - scipy_dia).nnz == 0

    def test_as_scipy_of_uninitialised_is_empty(self, shape):
        ndiag = 0
        base = dia.empty(shape[0], shape[1], ndiag)
        sci = base.as_scipy()
        assert len(sci.data) == 0
        assert len(sci.offsets) == 0

    def test_to_array_is_correct_result(self, data_diag):
        test_array = data_diag.to_array()
        assert isinstance(test_array, np.ndarray)
        # It's not enough to be accurate within a tolerance here - there's no
        # mathematics, so they should be _identical_.
        assert np.all(test_array == data_diag.as_scipy().toarray())


class TestFactoryMethods:
    def test_empty(self, shape, density):
        ndiag = int(shape[0] * shape[1] * density) or 1
        base = dia.empty(shape[0], shape[1], ndiag)
        sci = base.as_scipy(full=True)
        assert isinstance(base, dia.Dia)
        assert isinstance(sci, scipy.sparse.dia_matrix)
        assert base.shape == shape
        assert sci.data.shape == (ndiag, shape[1])
        assert sci.offsets.shape == (ndiag,)

    def test_zeros(self, shape):
        base = dia.zeros(shape[0], shape[1])
        sci = base.as_scipy()
        assert isinstance(base, dia.Dia)
        assert base.shape == shape
        assert sci.nnz == 0
        assert np.all(base.to_array() == 0)

    @pytest.mark.parametrize('dimension', [1, 5, 100])
    @pytest.mark.parametrize(
        'scale',
        [None, 2, -0.1, 1.5, 1.5+1j],
        ids=['none', 'int', 'negative', 'float', 'complex']
    )
    def test_identity(self, dimension, scale):
        # scale=None is testing that the default value returns the identity.
        base = (dia.identity(dimension) if scale is None
                else dia.identity(dimension, scale))
        sci = base.as_scipy()
        scipy_test = scipy.sparse.eye(dimension,
                                      dtype=np.complex128, format='dia')
        if scale is not None:
            scipy_test *= scale
        assert isinstance(base, dia.Dia)
        assert base.shape == (dimension, dimension)
        assert sci.nnz == dimension
        assert (sci - scipy_test).nnz == 0


    @pytest.mark.parametrize(['diagonals', 'offsets', 'shape'], [
        pytest.param([2j, 3, 5, 9], None, None, id='main diagonal'),
        pytest.param([1], None, None, id='1x1'),
        pytest.param([[0.2j, 0.3]], None, None, id='main diagonal list'),
        pytest.param([0.2j, 0.3], 2, None, id='superdiagonal'),
        pytest.param([0.2j, 0.3], -2, None, id='subdiagonal'),
        pytest.param([[0.2, 0.3, 0.4], [0.1, 0.9]], [-2, 3], None,
                     id='two diagonals'),
        pytest.param([1, 2, 3], 0, (3, 5), id='main wide'),
        pytest.param([1, 2, 3], 0, (5, 3), id='main tall'),
        pytest.param([[1, 2, 3], [4, 5]], [-1, -2], (4, 8), id='two wide sub'),
        pytest.param([[1, 2, 3, 4], [4, 5, 4j, 1j]], [1, 2], (4, 8),
                     id='two wide super'),
        pytest.param([[1, 2, 3], [4, 5]], [1, 2], (8, 4), id='two tall super'),
        pytest.param([[1, 2, 3, 4], [4, 5, 4j, 1j]], [-1, -2], (8, 4),
                     id='two tall sub'),
        pytest.param([[1, 2, 3], [4, 5, 6], [1, 2]], [1, -1, -2], (4, 4),
                     id='out of order'),
        pytest.param([[1, 2, 3], [4, 5, 6], [1, 2]], [1, 1, -2], (4, 4),
                     id='sum duplicates'),
    ])
    def test_diags(self, diagonals, offsets, shape):
        base = dia.diags(diagonals, offsets, shape)
        # Build numpy version test.
        if not isinstance(diagonals[0], list):
            diagonals = [diagonals]
        offsets = np.atleast_1d(offsets if offsets is not None else [0])
        if shape is None:
            size = len(diagonals[0]) + abs(offsets[0])
            shape = (size, size)
        test = np.zeros(shape, dtype=np.complex128)
        for diagonal, offset in zip(diagonals, offsets):
            test[np.where(np.eye(*shape, k=offset) == 1)] += diagonal
        assert isinstance(base, dia.Dia)
        assert base.shape == shape
        np.testing.assert_allclose(base.to_array(), test, rtol=1e-10)


    @pytest.mark.parametrize(['shape', 'position', 'value'], [
        pytest.param((1, 1), (0, 0), None, id='minimal'),
        pytest.param((10, 10), (5, 5), 1.j, id='on diagonal'),
        pytest.param((10, 10), (1, 5), 1., id='upper'),
        pytest.param((10, 10), (5, 1), 2., id='lower'),
        pytest.param((10, 1), (5, 0), None, id='column'),
        pytest.param((1, 10), (0, 5), -5.j, id='row'),
        pytest.param((10, 2), (5, 1), 1+2j, id='tall'),
        pytest.param((2, 10), (1, 5), 10, id='wide'),
    ])
    def test_one_element(self, shape, position, value):
        test = np.zeros(shape, dtype=np.complex128)
        if value is None:
            base = data.one_element_dia(shape, position)
            test[position] = 1.0+0.0j
        else:
            base = data.one_element_dia(shape, position, value)
            test[position] = value
        assert isinstance(base, data.Dia)
        assert base.shape == shape
        assert np.allclose(base.to_array(), test, atol=1e-10)

    @pytest.mark.parametrize(['shape', 'position', 'value'], [
        pytest.param((0, 0), (0, 0), None, id='zero shape'),
        pytest.param((10, -2), (5, 0), 1.j, id='neg shape'),
        pytest.param((10, 10), (10, 5), 1., id='outside'),
        pytest.param((10, 10), (5, -1), 2., id='outside neg'),
    ])
    def test_one_element_error(self, shape, position, value):
        with pytest.raises(ValueError) as exc:
            base = data.one_element_dia(shape, position, value)
        assert str(exc.value).startswith("Position of the elements"
                                         " out of bound: ")


def test_tidyup(data_diag):
    before = data_diag.to_array()
    sp_before = data_diag.as_scipy().toarray()
    largest = max(np.abs(before.real).max(), np.abs(before.imag).max())
    min_r = np.abs(before.real[np.abs(before.real) > 0]).min()
    min_i = np.abs(before.imag[np.abs(before.imag) > 0]).min()
    smallest = min(min_r, min_i)
    print(largest, smallest)
    if largest == smallest:
        return
    tol = (largest + smallest) / 2
    tidy = data.tidyup_dia(data_diag, tol, False)
    # Inplace=False, does not modify the original
    np.testing.assert_array_equal(data_diag.to_array(), before)
    np.testing.assert_array_equal(data_diag.as_scipy().toarray(), sp_before)
    # Is tidyup
    assert not np.allclose(tidy.to_array(), before)
    assert not np.allclose(tidy.as_scipy().toarray(), sp_before)

    data.tidyup_dia(data_diag, tol, True)
    assert not np.allclose(data_diag.to_array(), before)
    assert not np.allclose(data_diag.as_scipy().toarray(), sp_before)


def test_autotidyup():
    small = qeye(1) * 1e-5
    with CoreOptions(auto_tidyup_atol=1e-3):
        assert (small + small).tr() == 0
    with CoreOptions(auto_tidyup_atol=1e-3, auto_tidyup=False):
        assert (small + small).tr() == 2e-5



import numpy as np

from numpy.testing import assert_equal, assert_
import pytest
import functools
import itertools
import qutip
from qutip import (
    Qobj, identity, sigmax, to_super, to_choi, rand_super_bcsz, basis,
    tensor_contract, tensor_swap, num, QobjEvo, destroy, tensor,
    expand_operator
)


def test_tensor_contract_ident():
    qobj = identity([2, 3, 4])
    ans = 3 * identity([2, 4])

    assert_(ans == tensor_contract(qobj, (1, 4)))

    # Now try for superoperators.
    # For now, we just ensure the dims are correct.
    sqobj = to_super(qobj)
    correct_dims = [[[2, 4], [2, 4]], [[2, 4], [2, 4]]]
    assert_equal(correct_dims, tensor_contract(sqobj, (1, 4), (7, 10)).dims)


def case_tensor_contract_other(left, right, pairs,
                               expected_dims, expected_data):
    dat = np.arange(np.prod(left) * np.prod(right))
    dat = dat.reshape((np.prod(left), np.prod(right)))

    qobj = Qobj(dat, dims=[left, right])
    cqobj = tensor_contract(qobj, *pairs)
    assert_equal(cqobj.dims, expected_dims)
    assert_equal(cqobj.full(), expected_data)


def test_tensor_contract_other():
    case_tensor_contract_other(
        [2, 3], [3, 4], [(1, 2)],
        [[2], [4]],
        np.einsum('abbc', np.arange(2 * 3 * 3 * 4).reshape((2, 3, 3, 4)))
    )

    case_tensor_contract_other(
        [2, 3], [4, 3], [(1, 3)],
        [[2], [4]],
        np.einsum('abcb', np.arange(2 * 3 * 3 * 4).reshape((2, 3, 4, 3)))
    )

    case_tensor_contract_other(
        [2, 3], [4, 3], [(1, 3)],
        [[2], [4]],
        np.einsum('abcb', np.arange(2 * 3 * 3 * 4).reshape((2, 3, 4, 3)))
    )

    # Make non-rectangular outputs in a column-/row-symmetric fashion.
    big_dat = np.arange(2 * 3 * 2 * 3 * 2 * 3 * 2 * 3).reshape((2, 3) * 4)

    case_tensor_contract_other(
        [[2, 3], [2, 3]], [[2, 3], [2, 3]], [(0, 2)],
        [[[3], [3]], [[2, 3], [2, 3]]],
        np.einsum('ibidwxyz', big_dat).reshape((3 * 3, 3 * 2 * 3 * 2)))

    case_tensor_contract_other(
        [[2, 3], [2, 3]], [[2, 3], [2, 3]],
        [(0, 2), (5, 7)], [[[3], [3]], [[2], [2]]],
        # We separate einsum into two calls due to a bug internal to
        # einsum.
        np.einsum('ibidwy', np.einsum('abcdwjyj', big_dat)).reshape(3*3, 2*2))

    # Now we try collapsing in a way that's sensitive to column- and row-
    # stacking conventions.
    big_dat = np.arange(2 * 2 * 3 * 3 * 2 * 3 * 2 * 3)
    big_dat = big_dat.reshape((3, 3, 2, 2, 2, 3, 2, 3))
    # Note that the order of [2, 2] and [3, 3] is swapped on the left!
    big_dims = [[[2, 2], [3, 3]], [[2, 3], [2, 3]]]

    # Let's try a weird tensor contraction; this will likely never come up in
    # practice, but it should serve as a good canary for more reasonable
    # contractions.
    case_tensor_contract_other(
        big_dims[0], big_dims[1], [(0, 4)],
        [[[2], [3, 3]], [[3], [2, 3]]],
        np.einsum('abidwxiz', big_dat).reshape((2 * 3 * 3, 3 * 2 * 3)))


def case_tensor_swap(qobj, pairs, expected_dims, expected_data=None):
    sqobj = tensor_swap(qobj, *pairs)

    assert_equal(sqobj.dims, expected_dims)
    assert_equal(sqobj.full(), expected_data.full())


def test_tensor_swap_other():
    dims = (2, 3, 4, 5, 7)
    for dim in dims:
        S = to_super(rand_super_bcsz(dim))
        # Swapping the inner indices on a superoperator should give a Choi
        # matrix.
        J = to_choi(S)
        case_tensor_swap(S, [(1, 2)], [[[dim], [dim]], [[dim], [dim]]], J)


def test_tensor_qobjevo():
    N = 5
    t = 1.5
    left = QobjEvo([num(N),[destroy(N),"t"]])
    right = QobjEvo([identity(2),[sigmax(),"t"]])
    assert tensor(left, right)(t) == tensor(left(t), right(t))
    assert tensor(left, sigmax())(t) == tensor(left(t), sigmax())
    assert tensor(num(N), right)(t) == tensor(num(N), right(t))


def test_tensor_qobjevo_non_square():
    N = 5
    t = 1.5
    left = QobjEvo([basis(N, 0), [basis(N, 1), "t"]])
    right = QobjEvo([basis(2, 0).dag(), [basis(2, 1).dag(), "t"]])
    assert tensor(left, right)(t) == tensor(left(t), right(t))


def test_tensor_qobjevo_multiple():
    N = 5
    t = 1.5
    left = QobjEvo([basis(N, 0), [basis(N, 1), "t"]])
    center = QobjEvo([basis(2, 0).dag(), [basis(2, 1).dag(), "t"]])
    right = QobjEvo([sigmax()])
    as_QobjEvo = tensor(left, center, right)(t)
    as_Qobj = tensor(left(t), center(t), right(t))
    assert as_QobjEvo == as_Qobj


def test_tensor_and():
    N = 5
    t = 1.5
    sx = sigmax()
    evo = QobjEvo([num(N),[destroy(N),"t"]])
    assert tensor([sx, sx, sx]) == sx & sx & sx
    assert tensor(evo, sx)(t) == (evo & sx)(t)
    assert tensor(sx, evo)(t) == (sx & evo)(t)
    assert tensor(evo, evo)(t) == (evo & evo)(t)


def _permutation_id(permutation):
    return str(len(permutation)) + "-" + "".join(map(str, permutation))


def _tensor_with_entanglement(all_qubits, entangled, entangled_locations):
    """
    Create a full tensor product when a subspace component is already in an
    entangled state.  The locations in `all_qubits` which are the entangled
    points in the output are ignored and can take any value.

    For example,
        _tensor_with_entanglement([|a>, |b>, |c>, |d>], (|00> + |11>), [0, 2])
    should product a tensor product like (|0b0d> + |1b1d>), i.e. qubits 0 and 2
    in the final output are entangled, but the others are still separable.

    Parameters:
        all_qubits: list of kets --
            A list of separable parts to tensor together.  States that are in
            the locations referred to by `entangled_locations` are completely
            ignored.
        entangled: tensor-product ket -- the full entangled subspace
        entangled_locations: list of int --
            The locations that the qubits in the entangled subspace should be
            in in the final tensor-product space.
    """
    n_entangled = len(entangled.dims[0])
    n_separable = len(all_qubits) - n_entangled
    separable = all_qubits.copy()
    # Remove in reverse order so subsequent deletion locations don't change.
    for location in sorted(entangled_locations, reverse=True):
        del separable[location]
    # Can't separate out entangled states to pass to tensor in the right places
    # immediately, so tensor in at one end and then permute into place.
    out = qutip.tensor(*separable, entangled)
    permutation = list(range(n_separable))
    current_locations = range(n_separable, n_separable + n_entangled)
    # Sort to prevert later insertions changing previous locations.
    insertions = sorted(zip(entangled_locations, current_locations),
                        key=lambda x: x[0])
    for out_location, current_location in insertions:
        permutation.insert(out_location, current_location)
    return out.permute(permutation)


def _apply_permutation(permutation):
    """
    Permute the given permutation into the order denoted by its elements, i.e.
        out[0] = permutation[permutation[0]]
        out[1] = permutation[permutation[1]]
        ...

    This function is its own inverse.
    """
    out = [None] * len(permutation)
    for value, location in enumerate(permutation):
        out[location] = value
    return out


class Test_expand_operator:
    @pytest.mark.parametrize(
        'permutation',
        tuple(itertools.chain(*[
            itertools.permutations(range(k)) for k in [2, 3, 4]
        ])),
        ids=_permutation_id)
    def test_permutation_without_expansion(self, permutation):
        base = qutip.tensor([qutip.rand_unitary(2) for _ in permutation])
        test = expand_operator(base, [2] * len(permutation), permutation)
        expected = base.permute(_apply_permutation(permutation))
        np.testing.assert_allclose(test.full(), expected.full(), atol=1e-15)

    @pytest.mark.parametrize('n_targets', range(1, 5))
    def test_general_qubit_expansion(self, n_targets):
        # Test all permutations with the given number of targets.
        n_qubits = 5
        operation = qutip.rand_unitary([2]*n_targets)
        for targets in itertools.permutations(range(n_qubits), n_targets):
            expected = _tensor_with_entanglement([qutip.qeye(2)] * n_qubits,
                                                 operation, targets)
            test = expand_operator(operation, [2] * 5, targets)
            np.testing.assert_allclose(test.full(), expected.full(),
                                       atol=1e-15)

    def test_cnot_explicit(self):
        test = expand_operator(qutip.gates.cnot(), [2]*3, [2, 0]).full()
        expected = np.array([[1, 0, 0, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 0, 1, 0, 0],
                             [0, 0, 1, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 0, 0, 0, 1],
                             [0, 0, 0, 0, 1, 0, 0, 0],
                             [0, 1, 0, 0, 0, 0, 0, 0],
                             [0, 0, 0, 0, 0, 0, 1, 0],
                             [0, 0, 0, 1, 0, 0, 0, 0]])
        np.testing.assert_allclose(test, expected, atol=1e-15)

    @pytest.mark.parametrize('dimensions', [
        pytest.param([3, 4, 5], id="standard"),
        pytest.param([3, 3, 4, 4, 2], id="standard"),
        pytest.param([1, 2, 3], id="1D space"),
    ])
    def test_non_qubit_systems(self, dimensions):
        n_qubits = len(dimensions)
        for targets in itertools.permutations(range(n_qubits), 2):
            operators = [qutip.rand_unitary(dimension) if n in targets
                         else qutip.qeye(dimension)
                         for n, dimension in enumerate(dimensions)]
            expected = qutip.tensor(*operators)
            base_test = qutip.tensor(*[operators[x] for x in targets])
            test = expand_operator(base_test, dims=dimensions, targets=targets)
            assert test.dims == expected.dims
            np.testing.assert_allclose(test.full(), expected.full())



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

from qutip.core.data cimport CSR, Dense, Data, Dia

cpdef double one_csr(CSR matrix) except -1
cpdef double trace_csr(CSR matrix) except -1
cpdef double max_csr(CSR matrix) nogil
cpdef double frobenius_csr(CSR matrix) nogil
cpdef double l2_csr(CSR matrix) except -1 nogil

cpdef double frobenius_dense(Dense matrix) nogil
cpdef double l2_dense(Dense matrix) except -1 nogil

cpdef double one_dia(Dia matrix) except -1
cpdef double max_dia(Dia matrix) nogil
cpdef double frobenius_dia(Dia matrix) nogil
cpdef double l2_dia(Dia matrix) except -1 nogil

cpdef double frobenius_data(Data state) except -1

cdef inline int int_max(int a, int b) nogil:
    # Name collision between the ``max`` builtin and norm.max
    return b if b > a else a
